@@ -7,9 +7,9 @@ from .images import ImageStore, Resource
 
 def create_app(image_store):
     image_resource = Resource(image_store)
-    api = falcon.API()
-    api.add_route('/images', image_resource)
-    return api
+    app = falcon.App()
+    app.add_route('/images', image_resource)
+    return app
 
 
 def get_app():
