@@ -20,7 +20,7 @@ def falcon(body, headers):
     import falcon
 
     path = '/hello/{account_id}/test'
-    falcon_app = falcon.API('text/plain')
+    falcon_app = falcon.App('text/plain')
 
     # def ask(req, resp, params):
     #     params['answer'] = 42
@@ -70,7 +70,10 @@ def bottle(body, headers):
         user_agent = bottle.request.headers['User-Agent']  # NOQA
         limit = bottle.request.query.limit or '10'  # NOQA
 
-        return bottle.Response(body, headers=headers)
+        for header in headers.items():
+            bottle.response.set_header(*header)
+
+        return body
 
     return bottle.default_app()
 
@@ -96,59 +99,26 @@ def werkzeug(body, headers):
     return hello
 
 
-def cherrypy(body, headers):
-    import cherrypy
-
-    # Disable logging
-    cherrypy.config.update({'environment': 'embedded'})
-
-    class HelloResource(object):
-
-        exposed = True
-
-        def GET(self, account_id, test, limit=8):
-            user_agent = cherrypy.request.headers['User-Agent']  # NOQA
-            for name, value in headers.items():
-                cherrypy.response.headers[name] = value
-
-            return body
-
-    class Root(object):
-        pass
-
-    root = Root()
-    root.hello = HelloResource()
-
-    conf = {
-        '/': {
-            'request.dispatch': cherrypy.dispatch.MethodDispatcher(),
-        }
-    }
-
-    app = cherrypy.tree.mount(root, '/', conf)
-    return app
-
-
-# def wsme(body, headers):
-#     import wsme
-
-#     class HelloService(wsme.WSRoot):
-
-#         @wsme.expose(str, str)
-#         def hello(self, limit='10'):
-#             import pdb
-#             pdb.set_trace()
-#             return body
-
-#     ws = HelloService(protocols=['restjson'])
-#     return ws.wsgiapp()
-
-
 def pecan(body, headers):
-    import falcon.bench.nuts.nuts.app as nuts
+    import pecan
+    pecan.x_test_body = body
+    pecan.x_test_headers = headers
 
+    import falcon.bench.nuts.nuts.app as nuts
     sys.path.append(os.path.dirname(nuts.__file__))
     app = nuts.create()
     del sys.path[-1]
 
     return app
+
+
+def django(body, headers):
+    import django
+    django.x_test_body = body
+    django.x_test_headers = headers
+
+    from falcon.bench import dj
+    sys.path.append(os.path.dirname(dj.__file__))
+
+    from falcon.bench.dj.dj import wsgi
+    return wsgi.application
