@@ -8,7 +8,7 @@ _body = pecan.x_test_body
 _headers = pecan.x_test_headers
 
 
-class TestController(object):
+class TestController:
     def __init__(self, account_id):
         self.account_id = account_id
 
@@ -21,13 +21,13 @@ class TestController(object):
         return _body
 
 
-class HelloController(object):
+class HelloController:
     @expose()
     def _lookup(self, account_id, *remainder):
         return TestController(account_id), remainder
 
 
-class RootController(object):
+class RootController:
 
     @expose(content_type='text/plain')
     def index(self):
