@@ -14,7 +14,7 @@
 # limitations under the License.
 
 
-class Resource(object):
+class Resource:
 
     def on_get(self, req, resp, tenant_id, queue_name):
         pass
