@@ -14,14 +14,13 @@
 
 """HTTPError exception class."""
 
+from collections import OrderedDict
 import xml.etree.ElementTree as et
 
-try:
-    from collections import OrderedDict
-except ImportError:
-    OrderedDict = dict
-
-from falcon.util import json, uri
+from falcon.constants import MEDIA_JSON
+from falcon.util import uri
+from falcon.util.deprecation import deprecated
+from falcon.util.deprecation import deprecated_args
 
 
 class HTTPError(Exception):
@@ -33,31 +32,19 @@ class HTTPError(Exception):
     are supported by default.
 
     To customize the error presentation, implement a custom error
-    serializer and set it on the :class:`~.API` instance via
-    :meth:`~.API.set_error_serializer`.
+    serializer and set it on the :class:`~.App` instance via
+    :meth:`~.App.set_error_serializer`.
 
     To customize what data is passed to the serializer, subclass
     ``HTTPError`` and override the ``to_dict()`` method (``to_json()``
     is implemented via ``to_dict()``). To also support XML, override
     the ``to_xml()`` method.
 
-    Attributes:
-        status (str): HTTP status line, e.g. '748 Confounded by Ponies'.
-        has_representation (bool): Read-only property that determines
-            whether error details will be serialized when composing
-            the HTTP response. In ``HTTPError`` this property always
-            returns ``True``, but child classes may override it
-            in order to return ``False`` when an empty HTTP body is desired.
-
-            (See also: :class:`falcon.http_error.NoRepresentation`)
-
-        title (str): Error title to send to the client.
-        description (str): Description of the error to send to the client.
-        headers (dict): Extra headers to add to the response.
-        link (str): An href that the client can provide to the user for
-            getting help.
-        code (int): An internal application code that a user can reference when
-            requesting support for the error.
+    Note:
+        ``status`` is the only positional argument allowed, the other
+        arguments should be used as keyword only. Using them as positional
+        arguments will raise a deprecation warning and will result in an
+        error in a future version of falcon.
 
     Args:
         status (str): HTTP status code and text, such as "400 Bad Request"
@@ -87,11 +74,21 @@ class HTTPError(Exception):
         href (str): A URL someone can visit to find out more information
             (default ``None``). Unicode characters are percent-encoded.
         href_text (str): If href is given, use this as the friendly
-            title/description for the link (default 'API documentation
+            title/description for the link (default 'App documentation
             for this error').
         code (int): An internal code that customers can reference in their
             support request or to help them when searching for knowledge
             base articles related to this error (default ``None``).
+
+    Attributes:
+        status (str): HTTP status line, e.g. '748 Confounded by Ponies'.
+        title (str): Error title to send to the client.
+        description (str): Description of the error to send to the client.
+        headers (dict): Extra headers to add to the response.
+        link (str): An href that the client can provide to the user for
+            getting help.
+        code (int): An internal application code that a user can reference when
+            requesting support for the error.
     """
 
     __slots__ = (
@@ -103,6 +100,7 @@ class HTTPError(Exception):
         'code',
     )
 
+    @deprecated_args(allowed_positional=1)
     def __init__(self, status, title=None, description=None, headers=None,
                  href=None, href_text=None, code=None):
         self.status = status
@@ -128,7 +126,13 @@ class HTTPError(Exception):
     def __repr__(self):
         return '<%s: %s>' % (self.__class__.__name__, self.status)
 
-    @property
+    __str__ = __repr__
+
+    @property  # type: ignore
+    @deprecated(
+        'has_representation is deprecated and is currently unused by falcon',
+        is_property=True
+    )
     def has_representation(self):
         return True
 
@@ -163,22 +167,29 @@ class HTTPError(Exception):
 
         return obj
 
-    def to_json(self):
-        """Return a pretty-printed JSON representation of the error.
+    def to_json(self, handler=None):
+        """Return a JSON representation of the error.
+
+        Args:
+            handler: Handler object that will be used to serialize the representation of this
+                error to JSON. When not provided, a default handler using the builtin
+                JSON library will be used (default ``None``).
 
         Returns:
-            str: A JSON document for the error.
+            bytes: A JSON document for the error.
 
         """
 
         obj = self.to_dict(OrderedDict)
-        return json.dumps(obj, ensure_ascii=False)
+        if handler is None:
+            handler = _DEFAULT_JSON_HANDLER
+        return handler.serialize(obj, MEDIA_JSON)
 
     def to_xml(self):
         """Return an XML-encoded representation of the error.
 
         Returns:
-            str: An XML document for the error.
+            bytes: An XML document for the error.
 
         """
 
@@ -202,7 +213,12 @@ class HTTPError(Exception):
                 et.tostring(error_element, encoding='utf-8'))
 
 
-class NoRepresentation(object):
+# NOTE: initialized in falcon.media.json, that is always imported since Request/Respose
+# are imported by falcon init.
+_DEFAULT_JSON_HANDLER = None
+
+
+class NoRepresentation:
     """Mixin for ``HTTPError`` child classes that have no representation.
 
     This class can be mixed in when inheriting from ``HTTPError``, in order
@@ -219,14 +235,23 @@ class NoRepresentation(object):
         list when defining the child; otherwise, it will not override the
         `has_representation` property as expected.
 
+    Warning:
+        As of Falcon 3.0, this mixin class is no longer used since all Falcon
+        errors have a representation. This class is considered deprecated and
+        will be removed in a future release.
     """
 
-    @property
+    @property  # type: ignore
+    @deprecated(
+        'has_representation is deprecated and is currently unused by falcon. '
+        'The class NoRepresentation is deprecated and will be removed in a future release',
+        is_property=True
+    )
     def has_representation(self):
         return False
 
 
-class OptionalRepresentation(object):
+class OptionalRepresentation:
     """Mixin for ``HTTPError`` child classes that may have a representation.
 
     This class can be mixed in when inheriting from ``HTTPError`` in order
@@ -243,7 +268,17 @@ class OptionalRepresentation(object):
         list when defining the child; otherwise, it will not override the
         `has_representation` property as expected.
 
+    Warning:
+        As of Falcon 3.0, this mixin class is no longer used since all Falcon
+        errors have a representation. This class is considered deprecated and
+        will be removed in a future release.
     """
-    @property
+
+    @property  # type: ignore
+    @deprecated(
+        'has_representation is deprecated and is currently unused by falcon. '
+        'The class OptionalRepresentation is deprecated and will be removed in a future release',
+        is_property=True
+    )
     def has_representation(self):
-        return super(OptionalRepresentation, self).description is not None
+        return self.description is not None
