@@ -12,6 +12,10 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+"""HTTPStatus exception class."""
+
+from falcon.util import deprecated
+
 
 class HTTPStatus(Exception):
     """Represents a generic HTTP status.
@@ -20,27 +24,39 @@ class HTTPStatus(Exception):
     responder to short-circuit request processing in a manner similar
     to ``falcon.HTTPError``, but for non-error status codes.
 
-    Attributes:
-        status (str): HTTP status line, e.g. '748 Confounded by Ponies'.
-        headers (dict): Extra headers to add to the response.
-        body (str or unicode): String representing response content. If
-            Unicode, Falcon will encode as UTF-8 in the response.
-
     Args:
         status (str): HTTP status code and text, such as
             '748 Confounded by Ponies'.
         headers (dict): Extra headers to add to the response.
-        body (str or unicode): String representing response content. If
-            Unicode, Falcon will encode as UTF-8 in the response.
+        text (str): String representing response content. Falcon will encode
+            this value as UTF-8 in the response.
+        body (str): Deprecated alias to :attr:`text`. Will be removed in a future Falcon version.
+            :attr:`text` take precedence if provided.
+
+    Attributes:
+        status (str): HTTP status line, e.g. '748 Confounded by Ponies'.
+        headers (dict): Extra headers to add to the response.
+        text (str): String representing response content. Falcon will encode
+            this value as UTF-8 in the response.
+        body (str): Deprecated alias to :attr:`text`. Will be removed in a future Falcon version.
+
     """
 
     __slots__ = (
         'status',
         'headers',
-        'body'
+        'text'
     )
 
-    def __init__(self, status, headers=None, body=None):
+    def __init__(self, status, headers=None, text=None, body=None):
         self.status = status
         self.headers = headers
-        self.body = body
+        self.text = text if text is not None else body
+
+    @property  # type: ignore
+    @deprecated(
+        'Please use text instead.',
+        is_property=True
+    )
+    def body(self):
+        return self.text
