@@ -14,14 +14,22 @@
 
 """Default routing engine."""
 
+from collections import UserDict
+from inspect import iscoroutinefunction
 import keyword
 import re
 import textwrap
+from threading import Lock
 
 from falcon.routing import converters
-from falcon.routing.util import map_http_methods, set_default_responders
-from falcon.util.compat import UserDict
+from falcon.routing.util import map_http_methods
+from falcon.routing.util import set_default_responders
+from falcon.util.misc import is_python_func
+from falcon.util.sync import _should_wrap_non_coroutines
+from falcon.util.sync import wrap_sync_to_async
 
+if False:  # TODO: switch to TYPE_CHECKING once support for py3.5 is dropped
+    from typing import Any
 
 _TAB_STR = ' ' * 4
 _FIELD_PATTERN = re.compile(
@@ -37,17 +45,32 @@ _FIELD_PATTERN = re.compile(
 _IDENTIFIER_PATTERN = re.compile('[A-Za-z_][A-Za-z0-9_]*$')
 
 
-class CompiledRouter(object):
+class CompiledRouter:
     """Fast URI router which compiles its routing logic to Python code.
 
     Generally you do not need to use this router class directly, as an
-    instance is created by default when the falcon.API class is initialized.
+    instance is created by default when the falcon.App class is initialized.
 
     The router treats URI paths as a tree of URI segments and searches by
     checking the URI one segment at a time. Instead of interpreting the route
     tree for each look-up, it generates inlined, bespoke Python code to
     perform the search, then compiles that code. This makes the route
     processing quite fast.
+
+    The compilation process is delayed until the first use of the router (on the
+    first routed request) to reduce the time it takes to start the application.
+    This may noticeably delay the first response of the application when a large
+    number of routes have been added. When adding the last route
+    to the application a `compile` flag may be provided to force the router
+    to compile immediately, thus avoiding any delay for the first response.
+
+    Note:
+        When using a multi-threaded web server to host the application, it is
+        possible that multiple requests may be routed at the same time upon
+        startup. Therefore, the framework employs a lock to ensure that only a
+        single compilation of the decision tree is performed.
+
+    See also :meth:`.CompiledRouter.add_route`
     """
 
     __slots__ = (
@@ -60,6 +83,7 @@ class CompiledRouter(object):
         '_patterns',
         '_return_values',
         '_roots',
+        '_compile_lock',
     )
 
     def __init__(self):
@@ -77,9 +101,10 @@ class CompiledRouter(object):
         self._return_values = None
         self._roots = []
 
-        # NOTE(kgriffs): Call _compile() last since it depends on
-        # the variables above.
-        self._find = self._compile()
+        # NOTE(caselit): set _find to the delayed compile method to ensure that
+        # compile is called when the router is first used
+        self._find = self._compile_and_find
+        self._compile_lock = Lock()
 
     @property
     def options(self):
@@ -87,6 +112,10 @@ class CompiledRouter(object):
 
     @property
     def finder_src(self):
+        # NOTE(caselit): ensure that the router is actually compiled before
+        # returning the finder source, since the current value may be out of
+        # date
+        self.find('/')
         return self._finder_src
 
     def map_http_methods(self, resource, **kwargs):
@@ -119,7 +148,7 @@ class CompiledRouter(object):
         return map_http_methods(resource, suffix=kwargs.get('suffix', None))
 
     def add_route(self, uri_template, resource, **kwargs):
-        """Adds a route between a URI path template and a resource.
+        """Add a route between a URI path template and a resource.
 
         This method may be overridden to customize how a route is added.
 
@@ -139,17 +168,41 @@ class CompiledRouter(object):
                 Another class might use a suffixed responder to handle
                 a shortlink route in addition to the regular route for the
                 resource.
+            compile (bool): Optional flag that can be used to compile the
+                routing logic on this call. By default, :class:`.CompiledRouter`
+                delays compilation until the first request is routed. This may
+                introduce a noticeable amount of latency when handling the first
+                request, especially when the application implements a large
+                number of routes. Setting `compile` to ``True`` when the last
+                route is added ensures that the first request will not be
+                delayed in this case (defaults to ``False``).
+
+                Note:
+                    Always setting this flag to ``True`` may slow down the
+                    addition of new routes when hundreds of them are added at
+                    once. It is advisable to only set this flag to ``True`` when
+                    adding the final route.
         """
 
+        # NOTE(kgriffs): falcon.asgi.App injects this private kwarg; it is
+        #   only intended to be used internally.
+        asgi = kwargs.get('_asgi', False)
+
         method_map = self.map_http_methods(resource, **kwargs)
-        set_default_responders(method_map)
+
+        set_default_responders(method_map, asgi=asgi)
+
+        if asgi:
+            self._require_coroutine_responders(method_map)
+        else:
+            self._require_non_coroutine_responders(method_map)
 
         # NOTE(kgriffs): Fields may have whitespace in them, so sub
         # those before checking the rest of the URI template.
         if re.search(r'\s', _FIELD_PATTERN.sub('{FIELD}', uri_template)):
             raise ValueError('URI templates may not include whitespace.')
 
-        path = uri_template.strip('/').split('/')
+        path = uri_template.lstrip('/').split('/')
 
         used_names = set()
         for segment in path:
@@ -193,7 +246,12 @@ class CompiledRouter(object):
                 insert(new_node.children, path_index + 1)
 
         insert(self._roots)
-        self._find = self._compile()
+        # NOTE(caselit): when compile is True run the actual compile step, otherwise reset the
+        # _find, so that _compile will be called on the next find use
+        if kwargs.get('compile', False):
+            self._find = self._compile()
+        else:
+            self._find = self._compile_and_find
 
     def find(self, uri, req=None):
         """Search for a route that matches the given partial URI.
@@ -202,15 +260,15 @@ class CompiledRouter(object):
             uri(str): The requested path to route.
 
         Keyword Args:
-            req(Request): The Request object that will be passed to
-                the routed responder. Currently the value of this
-                argument is ignored by :class:`~.CompiledRouter`.
-                Routing is based solely on the path.
+            req: The :class:`falcon.Request` or :class:`falcon.asgi.Request`
+                object that will be passed to the routed responder. Currently
+                the value of this argument is ignored by
+                :class:`~.CompiledRouter`. Routing is based solely on the path.
 
         Returns:
             tuple: A 4-member tuple composed of (resource, method_map,
-                params, uri_template), or ``None`` if no route matches
-                the requested path.
+            params, uri_template), or ``None`` if no route matches
+            the requested path.
         """
 
         path = uri.lstrip('/').split('/')
@@ -227,8 +285,47 @@ class CompiledRouter(object):
     # Private
     # -----------------------------------------------------------------
 
+    def _require_coroutine_responders(self, method_map):
+        for method, responder in method_map.items():
+            # NOTE(kgriffs): We don't simply wrap non-async functions
+            #   since they likely peform relatively long blocking
+            #   operations that need to be explicitly made non-blocking
+            #   by the developer; raising an error helps highlight this
+            #   issue.
+            if not iscoroutinefunction(responder) and is_python_func(responder):
+                if _should_wrap_non_coroutines():
+                    def let(responder=responder):
+                        method_map[method] = wrap_sync_to_async(responder)
+
+                    let()
+
+                else:
+                    msg = (
+                        'The {} responder must be a non-blocking '
+                        'async coroutine (i.e., defined using async def) to '
+                        'avoid blocking the main request thread.'
+                    )
+                    msg = msg.format(responder)
+                    raise TypeError(msg)
+
+    def _require_non_coroutine_responders(self, method_map):
+        for method, responder in method_map.items():
+            # NOTE(kgriffs): We don't simply wrap non-async functions
+            #   since they likely peform relatively long blocking
+            #   operations that need to be explicitly made non-blocking
+            #   by the developer; raising an error helps highlight this
+            #   issue.
+
+            if iscoroutinefunction(responder):
+                msg = (
+                    'The {} responder must be a regular synchronous '
+                    'method to be used with a WSGI app.'
+                )
+                msg = msg.format(responder)
+                raise TypeError(msg)
+
     def _validate_template_segment(self, segment, used_names):
-        """Validates a single path segment of a URI template.
+        """Validate a single path segment of a URI template.
 
         1. Ensure field names are valid Python identifiers, since they
            will be passed as kwargs to responders.
@@ -265,12 +362,31 @@ class CompiledRouter(object):
                 raise ValueError(msg)
 
             name = field.group('cname')
-            if name and name not in self._converter_map:
-                msg = 'Unknown converter: "{0}"'.format(name)
-                raise ValueError(msg)
+            if name:
+                if name not in self._converter_map:
+                    msg = 'Unknown converter: "{0}"'.format(name)
+                    raise ValueError(msg)
+                try:
+                    self._instantiate_converter(self._converter_map[name], field.group('argstr'))
+                except Exception as e:
+                    msg = 'Cannot instantiate converter "{}"'.format(name)
+                    raise ValueError(msg) from e
 
-    def _generate_ast(self, nodes, parent, return_values, patterns, level=0, fast_return=True):
-        """Generates a coarse AST for the router."""
+    def _generate_ast(
+        self,
+        nodes: list,
+        parent,
+        return_values: list,
+        patterns: list,
+        params_stack: list,
+        level=0,
+        fast_return=True
+    ):
+        """Generate a coarse AST for the router."""
+        # NOTE(caselit): setting of the parameters in the params dict is delayed until
+        # a match has been found by adding them to the param_stack. This way superfluous
+        # parameters are not set to the params dict while descending on branches that
+        # ultimately do not match.
 
         # NOTE(kgriffs): Base case
         if not nodes:
@@ -301,7 +417,11 @@ class CompiledRouter(object):
 
                 fast_return = not found_var_nodes
 
+        construct = None  # type: Any
+        setter = None  # type: Any
+        original_params_stack = params_stack.copy()
         for node in nodes:
+            params_stack = original_params_stack.copy()
             if node.is_var:
                 if node.is_complex:
                     # NOTE(richardolsson): Complex nodes are nodes which
@@ -318,10 +438,13 @@ class CompiledRouter(object):
 
                     if node.var_converter_map:
                         parent.append_child(_CxPrefetchGroupsFromPatternMatch())
-                        parent = self._generate_conversion_ast(parent, node)
+                        parent = self._generate_conversion_ast(parent, node, params_stack)
 
                     else:
-                        parent.append_child(_CxSetParamsFromPatternMatch())
+                        construct = _CxVariableFromPatternMatch(len(params_stack) + 1)
+                        setter = _CxSetParamsFromDict(construct.dict_variable_name)
+                        params_stack.append(setter)
+                        parent.append_child(construct)
 
                 else:
                     # NOTE(kgriffs): Simple nodes just capture the entire path
@@ -343,15 +466,14 @@ class CompiledRouter(object):
                         converter_idx = len(self._converters)
                         self._converters.append(converter_obj)
 
-                        construct = _CxIfConverterField(
-                            field_name,
-                            converter_idx,
-                        )
+                        construct = _CxIfConverterField(len(params_stack) + 1, converter_idx)
+                        setter = _CxSetParamFromValue(field_name, construct.field_variable_name)
+                        params_stack.append(setter)
 
                         parent.append_child(construct)
                         parent = construct
                     else:
-                        parent.append_child(_CxSetParam(node.var_name, level))
+                        params_stack.append(_CxSetParamFromPath(node.var_name, level))
 
                     # NOTE(kgriffs): We don't allow multiple simple var nodes
                     # to exist at the same level, e.g.:
@@ -380,6 +502,7 @@ class CompiledRouter(object):
                 parent,
                 return_values,
                 patterns,
+                params_stack.copy(),
                 level + 1,
                 fast_return
             )
@@ -392,6 +515,8 @@ class CompiledRouter(object):
                 # the segments for the requested route; otherwise we could
                 # mistakenly match "/foo/23/bar" against "/foo/{id}".
                 construct = _CxIfPathLength('==', level + 1)
+                for params in params_stack:
+                    construct.append_child(params)
                 construct.append_child(_CxReturnValue(resource_idx))
                 parent.append_child(construct)
 
@@ -403,7 +528,9 @@ class CompiledRouter(object):
         if not found_simple and fast_return:
             parent.append_child(_CxReturnNone())
 
-    def _generate_conversion_ast(self, parent, node):
+    def _generate_conversion_ast(self, parent, node: 'CompiledRouterNode', params_stack: list):
+        construct = None  # type: Any
+        setter = None  # type: Any
         # NOTE(kgriffs): Unroll the converter loop into
         # a series of nested "if" constructs.
         for field_name, converter_name, converter_argstr in node.var_converter_map:
@@ -418,10 +545,9 @@ class CompiledRouter(object):
 
             parent.append_child(_CxSetFragmentFromField(field_name))
 
-            construct = _CxIfConverterField(
-                field_name,
-                converter_idx,
-            )
+            construct = _CxIfConverterField(len(params_stack) + 1, converter_idx)
+            setter = _CxSetParamFromValue(field_name, construct.field_variable_name)
+            params_stack.append(setter)
 
             parent.append_child(construct)
             parent = construct
@@ -429,12 +555,15 @@ class CompiledRouter(object):
         # NOTE(kgriffs): Add remaining fields that were not
         # converted, if any.
         if node.num_fields > len(node.var_converter_map):
-            parent.append_child(_CxSetParamsFromPatternMatchPrefetched())
+            construct = _CxVariableFromPatternMatchPrefetched(len(params_stack) + 1)
+            setter = _CxSetParamsFromDict(construct.dict_variable_name)
+            params_stack.append(setter)
+            parent.append_child(construct)
 
         return parent
 
     def _compile(self):
-        """Generates Python code for the entire routing tree.
+        """Generate Python code for the entire routing tree.
 
         The generated code is compiled and the resulting Python method
         is returned.
@@ -454,7 +583,8 @@ class CompiledRouter(object):
             self._roots,
             self._ast,
             self._return_values,
-            self._patterns
+            self._patterns,
+            params_stack=[]
         )
 
         src_lines.append(self._ast.src(0))
@@ -479,8 +609,27 @@ class CompiledRouter(object):
         src = '{0}({1})'.format(klass.__name__, argstr)
         return eval(src, {klass.__name__: klass})
 
+    def _compile_and_find(self, path, _return_values, _patterns, _converters, params):
+        """Compile the router, set the `_find` attribute and return its result.
 
-class CompiledRouterNode(object):
+        This method is set to the `_find` attribute to delay the compilation of the
+        router until it's used for the first time. Subsequent calls to `_find` will
+        be processed by the actual routing function.
+        This method must have the same signature as the function returned by the
+        :meth:`.CompiledRouter._compile`.
+        """
+        with self._compile_lock:
+            if self._find == self._compile_and_find:
+                # NOTE(caselit): replace the find with the result of the router compilation
+                self._find = self._compile()
+        # NOTE(caselit): return_values, patterns, converters are reset by the _compile
+        # method, so the updated ones must be used
+        return self._find(
+            path, self._return_values, self._patterns, self._converters, params
+        )
+
+
+class CompiledRouterNode:
     """Represents a single URI segment in a URI."""
 
     def __init__(self, raw_segment,
@@ -571,12 +720,12 @@ class CompiledRouterNode(object):
             assert self.is_var
 
     def matches(self, segment):
-        """Returns True if this node matches the supplied template segment."""
+        """Return True if this node matches the supplied template segment."""
 
         return segment == self.raw_segment
 
     def conflicts_with(self, segment):
-        """Returns True if this node conflicts with a given template segment."""
+        """Return True if this node conflicts with a given template segment."""
 
         # NOTE(kgriffs): This method assumes that the caller has already
         # checked if the segment matches. By definition, only unmatched
@@ -665,11 +814,12 @@ class ConverterDict(UserDict):
             )
 
 
-class CompiledRouterOptions(object):
+class CompiledRouterOptions:
     """Defines a set of configurable router options.
 
-    An instance of this class is exposed via :any:`API.router_options`
-    for configuring certain :py:class:`~.CompiledRouter` behaviors.
+    An instance of this class is exposed via :py:attr:`falcon.App.router_options`
+    and :py:attr:`falcon.asgi.App.router_options` for configuring certain
+    :py:class:`~.CompiledRouter` behaviors.
 
     Attributes:
         converters: Represents the collection of named
@@ -677,12 +827,12 @@ class CompiledRouterOptions(object):
             expressions. Adding additional converters is simply a
             matter of mapping an identifier to a converter class::
 
-                api.router_options.converters['mc'] = MyConverter
+                app.router_options.converters['mc'] = MyConverter
 
             The identifier can then be used to employ the converter
             within a URI template::
 
-                api.add_route('/{some_field:mc}', some_resource)
+                app.add_route('/{some_field:mc}', some_resource)
 
             Converter names may only contain ASCII letters, digits,
             and underscores, and must start with either a letter or
@@ -719,7 +869,7 @@ class CompiledRouterOptions(object):
 # --------------------------------------------------------------------
 
 
-class _CxParent(object):
+class _CxParent:
     def __init__(self):
         self._children = []
 
@@ -740,7 +890,7 @@ class _CxParent(object):
 
 class _CxIfPathLength(_CxParent):
     def __init__(self, comparison, length):
-        super(_CxIfPathLength, self).__init__()
+        super().__init__()
         self._comparison = comparison
         self._length = length
 
@@ -756,7 +906,7 @@ class _CxIfPathLength(_CxParent):
 
 class _CxIfPathSegmentLiteral(_CxParent):
     def __init__(self, segment_idx, literal):
-        super(_CxIfPathSegmentLiteral, self).__init__()
+        super().__init__()
         self._segment_idx = segment_idx
         self._literal = literal
 
@@ -772,7 +922,7 @@ class _CxIfPathSegmentLiteral(_CxParent):
 
 class _CxIfPathSegmentPattern(_CxParent):
     def __init__(self, segment_idx, pattern_idx, pattern_text):
-        super(_CxIfPathSegmentPattern, self).__init__()
+        super().__init__()
         self._segment_idx = segment_idx
         self._pattern_idx = pattern_idx
         self._pattern_text = pattern_text
@@ -793,21 +943,22 @@ class _CxIfPathSegmentPattern(_CxParent):
 
 
 class _CxIfConverterField(_CxParent):
-    def __init__(self, field_name, converter_idx):
-        super(_CxIfConverterField, self).__init__()
-        self._field_name = field_name
+    def __init__(self, unique_idx, converter_idx):
+        super().__init__()
         self._converter_idx = converter_idx
+        self._unique_idx = unique_idx
+        self.field_variable_name = 'field_value_{0}'.format(unique_idx)
 
     def src(self, indentation):
         lines = [
-            '{0}field_value = converters[{1}].convert(fragment)'.format(
+            '{0}{1} = converters[{2}].convert(fragment)'.format(
                 _TAB_STR * indentation,
+                self.field_variable_name,
                 self._converter_idx,
             ),
-            '{0}if field_value is not None:'.format(_TAB_STR * indentation),
-            "{0}params['{1}'] = field_value".format(
-                _TAB_STR * (indentation + 1),
-                self._field_name,
+            '{0}if {1} is not None:'.format(
+                _TAB_STR * indentation,
+                self.field_variable_name
             ),
             self._children_src(indentation + 1),
         ]
@@ -815,7 +966,7 @@ class _CxIfConverterField(_CxParent):
         return '\n'.join(lines)
 
 
-class _CxSetFragmentFromField(object):
+class _CxSetFragmentFromField:
     def __init__(self, field_name):
         self._field_name = field_name
 
@@ -826,7 +977,7 @@ class _CxSetFragmentFromField(object):
         )
 
 
-class _CxSetFragmentFromPath(object):
+class _CxSetFragmentFromPath:
     def __init__(self, segment_idx):
         self._segment_idx = segment_idx
 
@@ -837,33 +988,43 @@ class _CxSetFragmentFromPath(object):
         )
 
 
-class _CxSetParamsFromPatternMatch(object):
+class _CxVariableFromPatternMatch:
+    def __init__(self, unique_idx):
+        self._unique_idx = unique_idx
+        self.dict_variable_name = 'dict_match_{0}'.format(unique_idx)
+
     def src(self, indentation):
-        return '{0}params.update(match.groupdict())'.format(
-            _TAB_STR * indentation
+        return '{0}{1} = match.groupdict()'.format(
+            _TAB_STR * indentation,
+            self.dict_variable_name
         )
 
 
-class _CxSetParamsFromPatternMatchPrefetched(object):
+class _CxVariableFromPatternMatchPrefetched:
+    def __init__(self, unique_idx):
+        self._unique_idx = unique_idx
+        self.dict_variable_name = 'dict_groups_{0}'.format(unique_idx)
+
     def src(self, indentation):
-        return '{0}params.update(groups)'.format(
-            _TAB_STR * indentation
+        return '{0}{1} = groups'.format(
+            _TAB_STR * indentation,
+            self.dict_variable_name
         )
 
 
-class _CxPrefetchGroupsFromPatternMatch(object):
+class _CxPrefetchGroupsFromPatternMatch:
     def src(self, indentation):
         return '{0}groups = match.groupdict()'.format(
             _TAB_STR * indentation
         )
 
 
-class _CxReturnNone(object):
+class _CxReturnNone:
     def src(self, indentation):
         return '{0}return None'.format(_TAB_STR * indentation)
 
 
-class _CxReturnValue(object):
+class _CxReturnValue:
     def __init__(self, value_idx):
         self._value_idx = value_idx
 
@@ -874,7 +1035,7 @@ class _CxReturnValue(object):
         )
 
 
-class _CxSetParam(object):
+class _CxSetParamFromPath:
     def __init__(self, param_name, segment_idx):
         self._param_name = param_name
         self._segment_idx = segment_idx
@@ -885,3 +1046,27 @@ class _CxSetParam(object):
             self._param_name,
             self._segment_idx,
         )
+
+
+class _CxSetParamFromValue:
+    def __init__(self, param_name, field_value_name):
+        self._param_name = param_name
+        self._field_value_name = field_value_name
+
+    def src(self, indentation):
+        return "{0}params['{1}'] = {2}".format(
+            _TAB_STR * indentation,
+            self._param_name,
+            self._field_value_name,
+        )
+
+
+class _CxSetParamsFromDict:
+    def __init__(self, dict_value_name):
+        self._dict_value_name = dict_value_name
+
+    def src(self, indentation):
+        return '{0}params.update({1})'.format(
+            _TAB_STR * indentation,
+            self._dict_value_name,
+        )
