@@ -24,7 +24,16 @@ in the `falcon` module, and so must be explicitly imported::
 
 """
 
-import six
+from falcon.constants import PYPY
+try:
+    from falcon.cyutil.uri import (
+        decode as _cy_decode,
+        parse_query_string as _cy_parse_query_string
+    )
+except ImportError:
+    _cy_decode = None
+    _cy_parse_query_string = None
+
 
 # NOTE(kgriffs): See also RFC 3986
 _UNRESERVED = ('ABCDEFGHIJKLMNOPQRSTUVWXYZ'
@@ -38,6 +47,11 @@ _ALL_ALLOWED = _UNRESERVED + _DELIMITERS
 
 _HEX_DIGITS = '0123456789ABCDEFabcdef'
 
+# This map construction is based on urllib's implementation
+_HEX_TO_BYTE = {(a + b).encode(): bytes([int(a + b, 16)])
+                for a in _HEX_DIGITS
+                for b in _HEX_DIGITS}
+
 
 def _create_char_encoder(allowed_chars):
 
@@ -49,15 +63,12 @@ def _create_char_encoder(allowed_chars):
         else:
             encoded_char = '%{0:02X}'.format(code_point)
 
-        # NOTE(kgriffs): PY2 returns str from uri.encode, while
-        # PY3 returns a byte array.
-        key = chr(code_point) if six.PY2 else code_point
-        lookup[key] = encoded_char
+        lookup[code_point] = encoded_char
 
     return lookup.__getitem__
 
 
-def _create_str_encoder(is_value):
+def _create_str_encoder(is_value, check_is_escaped=False):
 
     allowed_chars = _UNRESERVED if is_value else _ALL_ALLOWED
     allowed_chars_plus_percent = allowed_chars + '%'
@@ -68,9 +79,14 @@ def _create_str_encoder(is_value):
         if not uri.rstrip(allowed_chars):
             return uri
 
-        if not uri.rstrip(allowed_chars_plus_percent):
+        if check_is_escaped and not uri.rstrip(allowed_chars_plus_percent):
             # NOTE(kgriffs): There's a good chance the string has already
             # been escaped. Do one more check to increase our certainty.
+            # NOTE(minesja): Per issue #1872, there's only certain situations
+            # in which we should check again (ex. location, content_location,
+            # append_link).In all other cases we should allow characters that
+            # could appear escaped to still be encoded (ex. '%' would be encoded
+            # as '%25).
             tokens = uri.split('%')
             for token in tokens[1:]:
                 hex_octet = token[:2]
@@ -93,15 +109,14 @@ def _create_str_encoder(is_value):
             # partially encoded, the caller will need to normalize it
             # before passing it in here.
 
-        # Convert to a byte array if it is not one already
-        if isinstance(uri, six.text_type):
-            uri = uri.encode('utf-8')
+        uri = uri.encode()
 
         # Use our map to encode each char and join the result into a new uri
         #
-        # PERF(kgriffs): map is faster than list comp on py27, but a tiny bit
-        # slower on py33. Since we are already much faster than urllib on
-        # py33, let's optimize for py27.
+        # PERF(kgriffs): map() is faster than list comp or generator comp on
+        # CPython 3 (tested on CPython 3.5 and 3.7). A list comprehension
+        # can be faster on PyPy3, but the difference is on the order of
+        # nanoseconds in that case, so we aren't going to worry about it.
         return ''.join(map(encode_char, uri))
 
     return encoder
@@ -121,10 +136,7 @@ Note:
     to use by assuming a sensible default of allowed characters.
 
 Args:
-    uri (str): URI or part of a URI to encode. If this is a wide
-        string (i.e., ``six.text_type``), it will be encoded to
-        a UTF-8 byte array and any multibyte sequences will
-        be percent-encoded as-is.
+    uri (str): URI or part of a URI to encode.
 
 Returns:
     str: An escaped version of `uri`, where all disallowed characters
@@ -132,7 +144,6 @@ Returns:
 
 """
 
-
 encode_value = _create_str_encoder(True)
 encode_value.name = 'encode_value'
 encode_value.__doc__ = """Encodes a value string according to RFC 3986.
@@ -153,10 +164,7 @@ Note:
 Args:
     uri (str): URI fragment to encode. It is assumed not to cross delimiter
         boundaries, and so any reserved URI delimiter characters
-        included in it will be escaped. If `value` is a wide
-        string (i.e., ``six.text_type``), it will be encoded to
-        a UTF-8 byte array and any multibyte sequences will
-        be percent-encoded as-is.
+        included in it will be percent-encoded.
 
 Returns:
     str: An escaped version of `uri`, where all disallowed characters
@@ -164,111 +172,149 @@ Returns:
 
 """
 
-if six.PY2:  # NOQA: C901 - Work around a bug in flake8 McCabe scoring
+encode_check_escaped = _create_str_encoder(False, True)
+encode_check_escaped.name = 'encode_check_escaped'
+encode_check_escaped.__doc__ = """Encodes a full or relative URI according to RFC 3986.
 
-    # This map construction is based on urllib
-    _HEX_TO_BYTE = dict((a + b, (chr(int(a + b, 16)), int(a + b, 16)))
-                        for a in _HEX_DIGITS
-                        for b in _HEX_DIGITS)
+RFC 3986 defines a set of "unreserved" characters as well as a
+set of "reserved" characters used as delimiters. This function escapes
+all other "disallowed" characters by percent-encoding them unless they
+appear to have been previously encoded. For example, ``'%26'`` will not be
+encoded again as it follows the format of an encoded value.
 
-    def decode(encoded_uri):
-        """Decodes percent-encoded characters in a URI or query string.
+Note:
+    This utility is faster in the average case than the similar
+    `quote` function found in ``urlib``. It also strives to be easier
+    to use by assuming a sensible default of allowed characters.
 
-        This function models the behavior of `urllib.unquote_plus`, but
-        is faster. It is also more robust, in that it will decode escaped
-        UTF-8 mutibyte sequences.
+Args:
+    uri (str): URI or part of a URI to encode.
 
-        Args:
-            encoded_uri (str): An encoded URI (full or partial).
+Returns:
+    str: An escaped version of `uri`, where all disallowed characters
+    have been percent-encoded.
 
-        Returns:
-            str: A decoded URL. Will be of type ``unicode`` on Python 2 IFF the
-            URL contained escaped non-ASCII characters, in which case
-            UTF-8 is assumed per RFC 3986.
+"""
 
-        """
+encode_value_check_escaped = _create_str_encoder(True, True)
+encode_value_check_escaped.name = 'encode_value_check_escaped'
+encode_value_check_escaped.__doc__ = """Encodes a value string according to RFC 3986.
 
-        decoded_uri = encoded_uri
+RFC 3986 defines a set of "unreserved" characters as well as a
+set of "reserved" characters used as delimiters. Disallowed characters
+are percent-encoded in a way that models ``urllib.parse.quote(safe="~")``
+unless they appear to have been previously encoded. For example, ``'%26'``
+will not be encoded again as it follows the format of an encoded value.
 
-        # PERF(kgriffs): Don't take the time to instantiate a new
-        # string unless we have to.
-        if '+' in decoded_uri:
-            decoded_uri = decoded_uri.replace('+', ' ')
+All reserved characters are lumped together into a single set of
+"delimiters", and everything in that set is escaped.
 
-        # Short-circuit if we can
-        if '%' not in decoded_uri:
-            return decoded_uri
+Note:
+    This utility is faster in the average case than the similar
+    `quote` function found in ``urlib``. It also strives to be easier
+    to use by assuming a sensible default of allowed characters.
 
-        # Convert to bytes because we are about to replace chars and we
-        # don't want Python to mistakenly interpret any high bits.
-        if not isinstance(decoded_uri, str):
-            # NOTE(kgriffs): Clients should never submit a URI that has
-            # unescaped non-ASCII chars in them, but just in case they
-            # do, let's encode in a non-lossy format.
-            decoded_uri = decoded_uri.encode('utf-8')
+Args:
+    uri (str): URI fragment to encode. It is assumed not to cross delimiter
+        boundaries, and so any reserved URI delimiter characters
+        included in it will be percent-encoded.
 
-        only_ascii = True
+Returns:
+    str: An escaped version of `uri`, where all disallowed characters
+    have been percent-encoded.
 
-        tokens = decoded_uri.split('%')
-        decoded_uri = tokens[0]
-        for token in tokens[1:]:
-            token_partial = token[:2]
-            try:
-                char, byte = _HEX_TO_BYTE[token_partial]
-            except KeyError:
-                char, byte = '%', 0
+"""
 
-            decoded_uri += char + (token[2:] if byte else token)
-            only_ascii = only_ascii and (byte <= 127)
 
-        # PERF(kgriffs): Only spend the time to do this if there
-        # were non-ascii bytes found in the string.
-        if not only_ascii:
-            decoded_uri = decoded_uri.decode('utf-8', 'replace')
+def _join_tokens_bytearray(tokens):
+    decoded_uri = bytearray(tokens[0])
+    for token in tokens[1:]:
+        token_partial = token[:2]
+        try:
+            decoded_uri += _HEX_TO_BYTE[token_partial] + token[2:]
+        except KeyError:
+            # malformed percentage like "x=%" or "y=%+"
+            decoded_uri += b'%' + token
 
+    # Convert back to str
+    return decoded_uri.decode('utf-8', 'replace')
+
+
+def _join_tokens_list(tokens):
+    decoded = tokens[:1]
+    # PERF(vytas): Do not copy list: a simple bool flag is fastest on PyPy JIT.
+    skip = True
+    for token in tokens:
+        if skip:
+            skip = False
+            continue
+
+        token_partial = token[:2]
+        try:
+            decoded.append(_HEX_TO_BYTE[token_partial] + token[2:])
+        except KeyError:
+            # malformed percentage like "x=%" or "y=%+"
+            decoded.append(b'%' + token)
+
+    # Convert back to str
+    return b''.join(decoded).decode('utf-8', 'replace')
+
+
+# PERF(vytas): The best method to join many byte strings depends on the Python
+#   implementation and many other factors such as the total string length,
+#   and the number of items to join.
+# Considering that CPython users are likely to choose the Cython implementation
+#   of decode anyway, we pick one best allrounder per platform:
+#   * On pure CPython, bytearray += often comes on top, also with the added
+#     benefit of being able to decode() off it directly.
+#   * On PyPy, b''.join(list) is the recommended approach, although it may
+#     narrowly lose to BytesIO on the extreme end.
+_join_tokens = _join_tokens_list if PYPY else _join_tokens_bytearray
+
+
+def decode(encoded_uri, unquote_plus=True):
+    """Decode percent-encoded characters in a URI or query string.
+
+    This function models the behavior of `urllib.parse.unquote_plus`,
+    albeit in a faster, more straightforward manner.
+
+    Args:
+        encoded_uri (str): An encoded URI (full or partial).
+
+    Keyword Arguments:
+        unquote_plus (bool): Set to ``False`` to retain any plus ('+')
+            characters in the given string, rather than converting them to
+            spaces (default ``True``). Typically you should set this
+            to ``False`` when decoding any part of a URI other than the
+            query string.
+
+    Returns:
+        str: A decoded URL. If the URL contains escaped non-ASCII
+        characters, UTF-8 is assumed per RFC 3986.
+
+    """
+
+    decoded_uri = encoded_uri
+
+    # PERF(kgriffs): Don't take the time to instantiate a new
+    # string unless we have to.
+    if '+' in decoded_uri and unquote_plus:
+        decoded_uri = decoded_uri.replace('+', ' ')
+
+    # Short-circuit if we can
+    if '%' not in decoded_uri:
         return decoded_uri
 
-else:
+    # NOTE(kgriffs): Clients should never submit a URI that has
+    # unescaped non-ASCII chars in them, but just in case they
+    # do, let's encode into a non-lossy format.
+    decoded_uri = decoded_uri.encode()
 
-    # This map construction is based on urllib
-    _HEX_TO_BYTE = dict(((a + b).encode(), bytes([int(a + b, 16)]))
-                        for a in _HEX_DIGITS
-                        for b in _HEX_DIGITS)
-
-    def decode(encoded_uri):
-        """Decodes percent-encoded characters in a URI or query string.
-
-        This function models the behavior of `urllib.parse.unquote_plus`,
-        albeit in a faster, more straightforward manner.
-
-        Args:
-            encoded_uri (str): An encoded URI (full or partial).
-
-        Returns:
-            str: A decoded URL. If the URL contains escaped non-ASCII
-            characters, UTF-8 is assumed per RFC 3986.
-
-        """
-
-        decoded_uri = encoded_uri
-
-        # PERF(kgriffs): Don't take the time to instantiate a new
-        # string unless we have to.
-        if '+' in decoded_uri:
-            decoded_uri = decoded_uri.replace('+', ' ')
-
-        # Short-circuit if we can
-        if '%' not in decoded_uri:
-            return decoded_uri
-
-        # NOTE(kgriffs): Clients should never submit a URI that has
-        # unescaped non-ASCII chars in them, but just in case they
-        # do, let's encode into a non-lossy format.
-        decoded_uri = decoded_uri.encode('utf-8')
-
-        # PERF(kgriffs): This was found to be faster than using
-        # a regex sub call or list comprehension with a join.
-        tokens = decoded_uri.split(b'%')
+    # PERF(kgriffs): This was found to be faster than using
+    # a regex sub call or list comprehension with a join.
+    tokens = decoded_uri.split(b'%')
+    # PERF(vytas): Just use in-place add for a low number of items:
+    if len(tokens) < 8:
         decoded_uri = tokens[0]
         for token in tokens[1:]:
             token_partial = token[:2]
@@ -281,9 +327,12 @@ else:
         # Convert back to str
         return decoded_uri.decode('utf-8', 'replace')
 
+    # NOTE(vytas): Decode percent-encoded bytestring fragments and join them
+    # back to a string using the platform-dependent method.
+    return _join_tokens(tokens)
 
-def parse_query_string(query_string, keep_blank_qs_values=False,
-                       parse_qs_csv=True):
+
+def parse_query_string(query_string, keep_blank=False, csv=True):
     """Parse a query string into a dict.
 
     Query string parameters are assumed to use standard form-encoding. Only
@@ -306,11 +355,11 @@ def parse_query_string(query_string, keep_blank_qs_values=False,
 
     Args:
         query_string (str): The query string to parse.
-        keep_blank_qs_values (bool): Set to ``True`` to return fields even if
+        keep_blank (bool): Set to ``True`` to return fields even if
             they do not have a value (default ``False``). For comma-separated
             values, this option also determines whether or not empty elements
             in the parsed list are retained.
-        parse_qs_csv: Set to ``False`` in order to disable splitting query
+        csv: Set to ``False`` in order to disable splitting query
             parameters on ``,`` (default ``True``). Depending on the user agent,
             encoding lists as multiple occurrences of the same parameter might
             be preferable. In this case, setting `parse_qs_csv` to ``False``
@@ -332,11 +381,10 @@ def parse_query_string(query_string, keep_blank_qs_values=False,
     is_encoded = '+' in query_string or '%' in query_string
 
     # PERF(kgriffs): This was found to be faster than using a regex, for
-    # both short and long query strings. Tested on both CPython 2.7 and 3.4,
-    # and on PyPy 2.3.
+    # both short and long query strings. Tested on CPython 3.4.
     for field in query_string.split('&'):
         k, _, v = field.partition('=')
-        if not (v or keep_blank_qs_values):
+        if not v and (not keep_blank or not k):
             continue
 
         # Note(steffgrez): Falcon first decode name parameter for handle
@@ -352,16 +400,7 @@ def parse_query_string(query_string, keep_blank_qs_values=False,
             # a list, or append the next value to the list.
             old_value = params[k]
 
-            if is_encoded:
-                v = decode(v)
-
-            if isinstance(old_value, list):
-                old_value.append(v)
-            else:
-                params[k] = [old_value, v]
-
-        else:
-            if parse_qs_csv and ',' in v:
+            if csv and ',' in v:
                 # NOTE(kgriffs): Falcon supports a more compact form of
                 # lists, in which the elements are comma-separated and
                 # assigned to a single param instance. If it turns out that
@@ -369,7 +408,39 @@ def parse_query_string(query_string, keep_blank_qs_values=False,
                 # point.
                 v = v.split(',')
 
-                if not keep_blank_qs_values:
+                if not keep_blank:
+                    # NOTE(kgriffs): Normalize the result in the case that
+                    # some elements are empty strings, such that the result
+                    # will be the same for 'foo=1,,3' as 'foo=1&foo=&foo=3'.
+                    additional_values = [decode(element) for element in v if element]
+                else:
+                    additional_values = [decode(element) for element in v]
+
+                if isinstance(old_value, list):
+                    old_value.extend(additional_values)
+                else:
+                    additional_values.insert(0, old_value)
+                    params[k] = additional_values
+
+            else:
+                if is_encoded:
+                    v = decode(v)
+
+                if isinstance(old_value, list):
+                    old_value.append(v)
+                else:
+                    params[k] = [old_value, v]
+
+        else:
+            if csv and ',' in v:
+                # NOTE(kgriffs): Falcon supports a more compact form of
+                # lists, in which the elements are comma-separated and
+                # assigned to a single param instance. If it turns out that
+                # very few people use this, it can be deprecated at some
+                # point.
+                v = v.split(',')
+
+                if not keep_blank:
                     # NOTE(kgriffs): Normalize the result in the case that
                     # some elements are empty strings, such that the result
                     # will be the same for 'foo=1,,3' as 'foo=1&foo=&foo=3'.
@@ -466,3 +537,21 @@ def unquote_string(quoted):
     else:
         return '\\'.join([q.replace('\\', '')
                           for q in tmp_quoted.split(r'\\')])
+
+
+# TODO(vytas): Restructure this in favour of a cleaner way to hoist the pure
+# Cython functions into this module.
+decode = _cy_decode or decode  # NOQA
+parse_query_string = _cy_parse_query_string or parse_query_string  # NOQA
+
+
+__all__ = [
+    'decode',
+    'encode',
+    'encode_value',
+    'encode_check_escaped',
+    'encode_value_check_escaped',
+    'parse_host',
+    'parse_query_string',
+    'unquote_string',
+]
