@@ -1,12 +1,12 @@
 import falcon
 
 
-class Things(object):
+class Things:
     def on_get(self, req, resp):
         pass
 
 
-api = application = falcon.API()
+api = application = falcon.App()
 api.add_route('/', Things())
 
 
