@@ -2,7 +2,7 @@ import io
 
 import pytest
 
-from falcon.request_helpers import BoundedStream
+from falcon.stream import BoundedStream
 
 
 @pytest.fixture
@@ -10,8 +10,8 @@ def bounded_stream():
     return BoundedStream(io.BytesIO(), 1024)
 
 
-def test_not_writeable(bounded_stream):
-    assert not bounded_stream.writeable()
+def test_not_writable(bounded_stream):
+    assert not bounded_stream.writable()
 
     with pytest.raises(IOError):
         bounded_stream.write(b'something something')
