@@ -2,14 +2,14 @@ import textwrap
 
 import pytest
 
-import falcon
 from falcon import testing
 from falcon.routing import DefaultRouter
 
+from _util import create_app  # NOQA
 
-@pytest.fixture
-def client():
-    return testing.TestClient(falcon.API())
+
+def client(asgi):
+    return testing.TestClient(create_app(asgi))
 
 
 @pytest.fixture
@@ -25,7 +25,7 @@ def router():
     router.add_route(
         '/repos/{org}/{repo}/commits', ResourceWithId(4))
     router.add_route(
-        u'/repos/{org}/{repo}/compare/{usr0}:{branch0}...{usr1}:{branch1}',
+        '/repos/{org}/{repo}/compare/{usr0}:{branch0}...{usr1}:{branch1}',
         ResourceWithId(5))
 
     router.add_route(
@@ -110,7 +110,7 @@ def router():
     return router
 
 
-class ResourceWithId(object):
+class ResourceWithId:
     def __init__(self, resource_id):
         self.resource_id = resource_id
 
@@ -118,10 +118,10 @@ class ResourceWithId(object):
         return 'ResourceWithId({})'.format(self.resource_id)
 
     def on_get(self, req, resp):
-        resp.body = self.resource_id
+        resp.text = self.resource_id
 
 
-class SpamConverter(object):
+class SpamConverter:
     def __init__(self, times, eggs=False):
         self._times = times
         self._eggs = eggs
@@ -244,13 +244,14 @@ def test_user_regression_special_chars(uri_template, path, expected_params):
 # =====================================================================
 
 
+@pytest.mark.parametrize('asgi', [True, False])
 @pytest.mark.parametrize('uri_template', [
     {},
     set(),
     object()
 ])
-def test_not_str(uri_template):
-    app = falcon.API()
+def test_not_str(asgi, uri_template):
+    app = create_app(asgi)
     with pytest.raises(TypeError):
         app.add_route(uri_template, ResourceWithId(-1))
 
@@ -367,7 +368,7 @@ def test_print_src(router):
 
     Example:
 
-        $ tox -e py27_debug -- -k test_print_src -s
+        $ tox -e py3_debug -- -k test_print_src -s
     """
     print('\n\n' + router.finder_src + '\n')
 
@@ -460,9 +461,11 @@ def test_converters_with_invalid_options(router, uri_template):
     # NOTE(kgriffs): Sanity-check that errors are properly bubbled up
     # when calling add_route(). Additional checks can be found
     # in test_uri_converters.py
-    with pytest.raises(ValueError):
+    with pytest.raises(ValueError, match='Cannot instantiate converter') as e:
         router.add_route(uri_template, ResourceWithId(1))
 
+    assert e.value.__cause__ is not None
+
 
 @pytest.mark.parametrize('uri_template', [
     '/foo/{bar:}',
@@ -650,3 +653,31 @@ def test_options_converters_invalid_name_on_update(router):
             'valid_name': SpamConverter,
             '7eleven': SpamConverter,
         })
+
+
+@pytest.fixture
+def param_router():
+    r = DefaultRouter()
+
+    r.add_route('/c/foo/{bar}/baz', ResourceWithId(1))
+    r.add_route('/c/{foo}/bar/other', ResourceWithId(2))
+    r.add_route('/c/foo/{a:int}-{b}/a', ResourceWithId(3))
+    r.add_route('/upload/{service}/auth/token', ResourceWithId(4))
+    r.add_route('/upload/youtube/{project_id}/share', ResourceWithId(5))
+    r.add_route('/x/y/{a}.{b}/z', ResourceWithId(6))
+    r.add_route('/x/{y}/o.o/w', ResourceWithId(7))
+    return r
+
+
+@pytest.mark.parametrize('route, expected, num', (
+    ('/c/foo/arg/baz', {'bar': 'arg'}, 1),
+    ('/c/foo/bar/other', {'foo': 'foo'}, 2),
+    ('/c/foo/42-7/baz', {'bar': '42-7'}, 1),
+    ('/upload/youtube/auth/token', {'service': 'youtube'}, 4),
+    ('/x/y/o.o/w', {'y': 'y'}, 7),
+))
+def test_params_in_non_taken_branches(param_router, route, expected, num):
+    resource, __, params, __ = param_router.find(route)
+
+    assert resource.resource_id == num
+    assert params == expected
