@@ -8,7 +8,7 @@ import requests
 import falcon
 
 
-class StorageEngine(object):
+class StorageEngine:
 
     def get_things(self, marker, limit):
         return [{'id': str(uuid.uuid4()), 'color': 'green'}]
@@ -22,15 +22,11 @@ class StorageError(Exception):
 
     @staticmethod
     def handle(req, resp, ex, params):
-        description = ("Sorry, couldn't write your thing to the "
-                       'database. It worked on my box.')
-
-        raise falcon.HTTPError(falcon.HTTP_725,
-                               'Database Error',
-                               description)
+        # TODO: Log the error, clean up, etc. before raising
+        raise falcon.HTTPInternalServerError()
 
 
-class SinkAdapter(object):
+class SinkAdapter:
 
     engines = {
         'ddg': 'https://duckduckgo.com',
@@ -42,12 +38,12 @@ class SinkAdapter(object):
         params = {'q': req.get_param('q', True)}
         result = requests.get(url, params=params)
 
-        resp.status = str(result.status_code) + ' ' + result.reason
+        resp.status = falcon.code_to_http_status(result.status_code)
         resp.content_type = result.headers['content-type']
-        resp.body = result.text
+        resp.text = result.text
 
 
-class AuthMiddleware(object):
+class AuthMiddleware:
 
     def process_request(self, req, resp):
         token = req.get_header('Authorization')
@@ -59,40 +55,42 @@ class AuthMiddleware(object):
             description = ('Please provide an auth token '
                            'as part of the request.')
 
-            raise falcon.HTTPUnauthorized('Auth token required',
-                                          description,
-                                          challenges,
+            raise falcon.HTTPUnauthorized(title='Auth token required',
+                                          description=description,
+                                          challenges=challenges,
                                           href='http://docs.example.com/auth')
 
         if not self._token_is_valid(token, account_id):
             description = ('The provided auth token is not valid. '
                            'Please request a new token and try again.')
 
-            raise falcon.HTTPUnauthorized('Authentication required',
-                                          description,
-                                          challenges,
+            raise falcon.HTTPUnauthorized(title='Authentication required',
+                                          description=description,
+                                          challenges=challenges,
                                           href='http://docs.example.com/auth')
 
     def _token_is_valid(self, token, account_id):
         return True  # Suuuuuure it's valid...
 
 
-class RequireJSON(object):
+class RequireJSON:
 
     def process_request(self, req, resp):
         if not req.client_accepts_json:
             raise falcon.HTTPNotAcceptable(
-                'This API only supports responses encoded as JSON.',
+                title='406 Not Acceptable',
+                description='This API only supports responses encoded as JSON.',
                 href='http://docs.examples.com/api/json')
 
         if req.method in ('POST', 'PUT'):
             if 'application/json' not in req.content_type:
                 raise falcon.HTTPUnsupportedMediaType(
-                    'This API only supports requests encoded as JSON.',
+                    title='415 Unsupported Media Type',
+                    description='This API only supports requests encoded as JSON.',
                     href='http://docs.examples.com/api/json')
 
 
-class JSONTranslator(object):
+class JSONTranslator:
 
     def process_request(self, req, resp):
         # req.stream corresponds to the WSGI wsgi.input environ variable,
@@ -105,24 +103,25 @@ class JSONTranslator(object):
 
         body = req.bounded_stream.read()
         if not body:
-            raise falcon.HTTPBadRequest('Empty request body',
-                                        'A valid JSON document is required.')
+            raise falcon.HTTPBadRequest(title='Empty request body',
+                                        description='A valid JSON document is required.')
 
         try:
             req.context.doc = json.loads(body.decode('utf-8'))
 
         except (ValueError, UnicodeDecodeError):
-            raise falcon.HTTPError(falcon.HTTP_753,
-                                   'Malformed JSON',
-                                   'Could not decode the request body. The '
-                                   'JSON was incorrect or not encoded as '
-                                   'UTF-8.')
+            description = ('Could not decode the request body. The '
+                           'JSON was incorrect or not encoded as '
+                           'UTF-8.')
 
-    def process_response(self, req, resp, resource):
-        if not hasattr(req.context, 'result'):
+            raise falcon.HTTPBadRequest(title='Malformed JSON',
+                                        description=description)
+
+    def process_response(self, req, resp, resource, req_succeeded):
+        if not hasattr(resp.context, 'result'):
             return
 
-        resp.body = json.dumps(req.context.result)
+        resp.text = json.dumps(resp.context.result)
 
 
 def max_body(limit):
@@ -134,12 +133,12 @@ def max_body(limit):
                    'exceed ' + str(limit) + ' bytes in length.')
 
             raise falcon.HTTPPayloadTooLarge(
-                'Request body is too large', msg)
+                title='Request body is too large', description=msg)
 
     return hook
 
 
-class ThingsResource(object):
+class ThingsResource:
 
     def __init__(self, db):
         self.db = db
@@ -159,15 +158,15 @@ class ThingsResource(object):
                            'We appreciate your patience.')
 
             raise falcon.HTTPServiceUnavailable(
-                'Service Outage',
-                description,
-                30)
+                title='Service Outage',
+                description=description,
+                retry_after=30)
 
-        # An alternative way of doing DRY serialization would be to
-        # create a custom class that inherits from falcon.Request. This
-        # class could, for example, have an additional 'doc' property
-        # that would serialize to JSON under the covers.
-        req.context.result = result
+        # NOTE: Normally you would use resp.media for this sort of thing;
+        # this example serves only to demonstrate how the context can be
+        # used to pass arbitrary values between middleware components,
+        # hooks, and resources.
+        resp.context.result = result
 
         resp.set_header('Powered-By', 'Falcon')
         resp.status = falcon.HTTP_200
@@ -178,8 +177,8 @@ class ThingsResource(object):
             doc = req.context.doc
         except AttributeError:
             raise falcon.HTTPBadRequest(
-                'Missing thing',
-                'A thing must be submitted in the request body.')
+                title='Missing thing',
+                description='A thing must be submitted in the request body.')
 
         proper_thing = self.db.add_thing(doc)
 
@@ -188,7 +187,7 @@ class ThingsResource(object):
 
 
 # Configure your WSGI server to load "things.app" (app is a WSGI callable)
-app = falcon.API(middleware=[
+app = falcon.App(middleware=[
     AuthMiddleware(),
     RequireJSON(),
     JSONTranslator(),
@@ -198,7 +197,7 @@ db = StorageEngine()
 things = ThingsResource(db)
 app.add_route('/{user_id}/things', things)
 
-# If a responder ever raised an instance of StorageError, pass control to
+# If a responder ever raises an instance of StorageError, pass control to
 # the given handler.
 app.add_error_handler(StorageError, StorageError.handle)
 
