@@ -5,19 +5,20 @@ import pytest
 
 import falcon
 from falcon import testing
-from falcon.util import compat
+
+from _util import create_app  # NOQA
 
 
 SAMPLE_BODY = testing.rand_string(0, 128 * 1024)
 
 
 @pytest.fixture
-def client():
-    app = falcon.API()
+def client(asgi):
+    app = create_app(asgi)
     return testing.TestClient(app)
 
 
-class XmlResource(object):
+class XmlResource:
     def __init__(self, content_type):
         self.content_type = content_type
 
@@ -25,7 +26,7 @@ class XmlResource(object):
         resp.set_header('content-type', self.content_type)
 
 
-class HeaderHelpersResource(object):
+class HeaderHelpersResource:
 
     def __init__(self, last_modified=None):
         if last_modified is not None:
@@ -38,7 +39,7 @@ class HeaderHelpersResource(object):
         resp.cache_control = ['no-store']
 
     def on_get(self, req, resp):
-        resp.body = '{}'
+        resp.text = '{}'
         resp.content_type = 'x-falcon/peregrine'
         resp.cache_control = [
             'public', 'private', 'no-cache', 'no-store', 'must-revalidate',
@@ -71,6 +72,7 @@ class HeaderHelpersResource(object):
         assert resp.get_header('x-client-should-never-see-this') == 'abc'
         resp.delete_header('x-client-should-never-see-this')
 
+        self.req = req
         self.resp = resp
 
     def on_head(self, req, resp):
@@ -107,10 +109,10 @@ class HeaderHelpersResource(object):
         self.resp = resp
 
 
-class LocationHeaderUnicodeResource(object):
+class LocationHeaderUnicodeResource:
 
-    URL1 = u'/\u00e7runchy/bacon'
-    URL2 = u'ab\u00e7' if compat.PY3 else 'ab\xc3\xa7'
+    URL1 = '/\u00e7runchy/bacon'
+    URL2 = 'ab\u00e7'
 
     def on_get(self, req, resp):
         resp.location = self.URL1
@@ -121,52 +123,63 @@ class LocationHeaderUnicodeResource(object):
         resp.content_location = self.URL1
 
 
-class UnicodeHeaderResource(object):
+class UnicodeHeaderResource:
+
+    def on_connect(self, req, resp):
+        # A way to CONNECT with people.
+        resp.set_header('X-Clinking-Beer-Mugs', '🍺')
 
     def on_get(self, req, resp):
         resp.set_headers([
-            (u'X-auTH-toKEN', 'toomanysecrets'),
-            ('Content-TYpE', u'application/json'),
-            (u'X-symBOl', u'@'),
+            ('X-auTH-toKEN', 'toomanysecrets'),
+            ('Content-TYpE', 'application/json'),
+            ('X-symbOl', '@'),
+        ])
+
+    def on_patch(self, req, resp):
+        resp.set_headers([
+            ('X-Thing', '\x01\x02\xff'),
         ])
 
     def on_post(self, req, resp):
         resp.set_headers([
-            (u'X-symb\u00F6l', 'thing'),
+            ('X-symb\u00F6l', 'thing'),
         ])
 
     def on_put(self, req, resp):
         resp.set_headers([
-            ('X-Thing', u'\u00FF'),
+            ('X-Thing', '\u00FF'),
         ])
 
 
-class VaryHeaderResource(object):
+class VaryHeaderResource:
 
     def __init__(self, vary):
         self.vary = vary
 
     def on_get(self, req, resp):
-        resp.body = '{}'
+        resp.text = '{}'
         resp.vary = self.vary
 
 
-class LinkHeaderResource(object):
+class LinkHeaderResource:
 
     def __init__(self):
         self._links = []
 
-    def add_link(self, *args, **kwargs):
+    def append_link(self, *args, **kwargs):
         self._links.append((args, kwargs))
 
     def on_get(self, req, resp):
-        resp.body = '{}'
+        resp.text = '{}'
 
+        append_link = None
         for args, kwargs in self._links:
-            resp.add_link(*args, **kwargs)
+            append_link = resp.append_link if append_link is resp.add_link else resp.add_link
+            append_link(*args, **kwargs)
 
 
-class AppendHeaderResource(object):
+class AppendHeaderResource:
 
     def on_get(self, req, resp):
         resp.append_header('X-Things', 'thing-1')
@@ -187,7 +200,7 @@ class AppendHeaderResource(object):
         resp.append_header('seT-cookie', c2)
 
 
-class RemoveHeaderResource(object):
+class RemoveHeaderResource:
     def __init__(self, with_double_quotes):
         self.with_double_quotes = with_double_quotes
 
@@ -205,7 +218,17 @@ class RemoveHeaderResource(object):
         resp.downloadable_as = None
 
 
-class ContentLengthHeaderResource(object):
+class DownloadableResource:
+    def __init__(self, filename):
+        self.filename = filename
+
+    def on_get(self, req, resp):
+        resp.text = 'Hello, World!\n'
+        resp.content_type = falcon.MEDIA_TEXT
+        resp.downloadable_as = self.filename
+
+
+class ContentLengthHeaderResource:
 
     def __init__(self, content_length, body=None, data=None):
         self._content_length = content_length
@@ -213,11 +236,10 @@ class ContentLengthHeaderResource(object):
         self._data = data
 
     def on_get(self, req, resp):
-        # NOTE(kgriffs): Use stream_len for now to cover the deprecated alias
-        resp.stream_len = self._content_length
+        resp.content_length = self._content_length
 
         if self._body:
-            resp.body = self._body
+            resp.text = self._body
 
         if self._data:
             resp.data = self._data
@@ -226,7 +248,7 @@ class ContentLengthHeaderResource(object):
         resp.content_length = self._content_length
 
 
-class ExpiresHeaderResource(object):
+class ExpiresHeaderResource:
 
     def __init__(self, expires):
         self._expires = expires
@@ -235,7 +257,34 @@ class ExpiresHeaderResource(object):
         resp.expires = self._expires
 
 
-class TestHeaders(object):
+class CustomHeaders:
+    def items(self):
+        return [('test-header', 'test-value')]
+
+
+class CustomHeadersNotCallable:
+    def __init__(self):
+        self.items = {'test-header': 'test-value'}
+
+
+class CustomHeadersResource:
+    def on_get(self, req, resp):
+        headers = CustomHeaders()
+        resp.set_headers(headers)
+
+    def on_post(self, req, resp):
+        resp.set_headers(CustomHeadersNotCallable())
+
+
+class HeadersDebugResource:
+    def on_get(self, req, resp):
+        resp.media = {key.lower(): value for key, value in req.headers.items()}
+
+    def on_get_header(self, req, resp, header):
+        resp.media = {header.lower(): req.get_header(header)}
+
+
+class TestHeaders:
 
     def test_content_length(self, client):
         resource = testing.SimpleTestResource(body=SAMPLE_BODY)
@@ -290,6 +339,10 @@ class TestHeaders(object):
         value = req.get_header('X-Not-Found', default='some-value')
         assert value == 'some-value'
 
+        # Excercise any result caching and associated abuse mitigations
+        for i in range(10000):
+            assert req.get_header('X-Not-Found-{0}'.format(i)) is None
+
     @pytest.mark.parametrize('with_double_quotes', [True, False])
     def test_unset_header(self, client, with_double_quotes):
         client.app.add_route('/', RemoveHeaderResource(with_double_quotes))
@@ -310,7 +363,7 @@ class TestHeaders(object):
         except falcon.HTTPMissingHeader as ex:
             assert isinstance(ex, falcon.HTTPBadRequest)
             assert ex.title == 'Missing header value'
-            expected_desc = 'The X-Not-Found header is required.'
+            expected_desc = 'The "X-Not-Found" header is required.'
             assert ex.description == expected_desc
 
     @pytest.mark.parametrize('status', (falcon.HTTP_204, falcon.HTTP_304))
@@ -342,6 +395,9 @@ class TestHeaders(object):
 
         client.simulate_get(headers=resource.captured_req.headers)
 
+        # Validate that the property has been cached
+        assert resource.captured_req.headers is resource.captured_req.headers
+
         # Compare the request HTTP headers with the original headers
         for name, expected_value in request_headers.items():
             actual_value = resource.captured_req.get_header(name)
@@ -371,31 +427,33 @@ class TestHeaders(object):
         resource = testing.SimpleTestResource(body='Hello world!')
         self._check_header(client, resource, 'Content-Type', falcon.DEFAULT_MEDIA_TYPE)
 
+    @pytest.mark.parametrize('asgi', [True, False])
     @pytest.mark.parametrize('content_type,body', [
-        ('text/plain; charset=UTF-8', u'Hello Unicode! \U0001F638'),
+        ('text/plain; charset=UTF-8', 'Hello Unicode! \U0001F638'),
         # NOTE(kgriffs): This only works because the client defaults to
         # ISO-8859-1 IFF the media type is 'text'.
         ('text/plain', 'Hello ISO-8859-1!'),
     ])
-    def test_override_default_media_type(self, client, content_type, body):
-        client.app = falcon.API(media_type=content_type)
+    def test_override_default_media_type(self, asgi, client, content_type, body):
+        client.app = create_app(asgi=asgi, media_type=content_type)
         client.app.add_route('/', testing.SimpleTestResource(body=body))
         result = client.simulate_get()
 
         assert result.text == body
         assert result.headers['Content-Type'] == content_type
 
-    def test_override_default_media_type_missing_encoding(self, client):
-        body = u'{"msg": "Hello Unicode! \U0001F638"}'
+    @pytest.mark.parametrize('asgi', [True, False])
+    def test_override_default_media_type_missing_encoding(self, asgi, client):
+        body = '{"msg": "Hello Unicode! \U0001F638"}'
 
-        client.app = falcon.API(media_type='application/json')
+        client.app = create_app(asgi=asgi, media_type='application/json')
         client.app.add_route('/', testing.SimpleTestResource(body=body))
         result = client.simulate_get()
 
         assert result.content == body.encode('utf-8')
-        assert isinstance(result.text, compat.text_type)
+        assert isinstance(result.text, str)
         assert result.text == body
-        assert result.json == {u'msg': u'Hello Unicode! \U0001F638'}
+        assert result.json == {'msg': 'Hello Unicode! \U0001F638'}
 
     def test_response_header_helpers_on_get(self, client):
         last_modified = datetime(2013, 1, 1, 10, 30, 30)
@@ -438,7 +496,7 @@ class TestHeaders(object):
         assert resp.content_range == content_range
         assert result.headers['Content-Range'] == content_range
 
-        resp.content_range = (1, 499, 10 * 1024, u'bytes')
+        resp.content_range = (1, 499, 10 * 1024, 'bytes')
         assert isinstance(resp.content_range, str)
         assert resp.content_range == 'bytes 1-499/10240'
 
@@ -455,6 +513,63 @@ class TestHeaders(object):
             hist[name] += 1
             assert 1 == hist[name]
 
+    @pytest.mark.parametrize('filename,expected', [
+        ('report.csv', 'attachment; filename="report.csv"'),
+        ('Hello World.txt', 'attachment; filename="Hello World.txt"'),
+        (
+            'Bold Digit 𝟏.txt',
+            'attachment; filename=Bold_Digit_1.txt; '
+            "filename*=UTF-8''Bold%20Digit%20%F0%9D%9F%8F.txt",
+        ),
+        (
+            'Ångström unit.txt',
+            'attachment; filename=A_ngstro_m_unit.txt; '
+            "filename*=UTF-8''%C3%85ngstr%C3%B6m%20unit.txt",
+        ),
+        ('one,two.txt', 'attachment; filename="one,two.txt"'),
+        (
+            '½,²⁄₂.txt',
+            'attachment; filename=1_2_2_2.txt; '
+            "filename*=UTF-8''%C2%BD%2C%C2%B2%E2%81%84%E2%82%82.txt"
+        ),
+        ('[foo] @ bar.txt', 'attachment; filename="[foo] @ bar.txt"'),
+        (
+            '[fòó]@bàr,bäz.txt',
+            'attachment; filename=_fo_o___ba_r_ba_z.txt; '
+            "filename*=UTF-8''%5Bf%C3%B2%C3%B3%5D%40b%C3%A0r%2Cb%C3%A4z.txt"
+        ),
+    ])
+    def test_content_disposition_header(self, client, filename, expected):
+        resource = DownloadableResource(filename)
+        client.app.add_route('/', resource)
+        resp = client.simulate_get()
+
+        assert resp.status_code == 200
+        assert resp.headers['Content-Disposition'] == expected
+
+    def test_request_latin1_headers(self, client):
+        client.app.add_route('/headers', HeadersDebugResource())
+        client.app.add_route('/headers/{header}', HeadersDebugResource(), suffix='header')
+
+        headers = {
+            'User-Agent': 'Mosaic/0.9',
+            'X-Latin1-Header': 'Förmånsrätt',
+            'X-Size': 'groß',
+        }
+        resp = client.simulate_get('/headers', headers=headers)
+        assert resp.status_code == 200
+        assert resp.json == {
+            'host': 'falconframework.org',
+            'user-agent': 'Mosaic/0.9',
+            'x-latin1-header': 'Förmånsrätt',
+            'x-size': 'groß',
+        }
+
+        resp = client.simulate_get('/headers/X-Latin1-Header', headers=headers)
+        assert resp.json == {'x-latin1-header': 'Förmånsrätt'}
+        resp = client.simulate_get('/headers/X-Size', headers=headers)
+        assert resp.json == {'x-size': 'groß'}
+
     def test_unicode_location_headers(self, client):
         client.app.add_route('/', LocationHeaderUnicodeResource())
 
@@ -467,7 +582,7 @@ class TestHeaders(object):
         assert result.headers['Content-Location'] == '/%C3%A7runchy/bacon'
         assert result.headers['Location'] == 'ab%C3%A7'
 
-    def test_unicode_headers_convertable(self, client):
+    def test_unicode_headers_contain_only_ascii(self, client):
         client.app.add_route('/', UnicodeHeaderResource())
 
         result = client.simulate_get('/')
@@ -476,14 +591,42 @@ class TestHeaders(object):
         assert result.headers['X-Auth-Token'] == 'toomanysecrets'
         assert result.headers['X-Symbol'] == '@'
 
-    @pytest.mark.skipif(compat.PY3, reason='Test only applies to Python 2')
-    def test_unicode_headers_not_convertable(self, client):
-        client.app.add_route('/', UnicodeHeaderResource())
-        with pytest.raises(UnicodeEncodeError):
-            client.simulate_post('/')
+    @pytest.mark.parametrize('method', ['CONNECT', 'PATCH', 'POST', 'PUT'])
+    def test_unicode_headers_contain_non_ascii(self, method, client):
+        app = client.app
+        app.add_route('/', UnicodeHeaderResource())
 
-        with pytest.raises(UnicodeEncodeError):
-            client.simulate_put('/')
+        if method == 'CONNECT':
+            # NOTE(vytas): Response headers cannot be encoded to Latin-1.
+            if not app._ASGI:
+                pytest.skip('wsgiref.validate sees no evil here')
+
+            # NOTE(vytas): Shouldn't this result in an HTTP 500 instead of
+            #   bubbling up a ValueError to the app server?
+            with pytest.raises(ValueError):
+                client.simulate_request(method, '/')
+        elif method == 'PUT':
+            # NOTE(vytas): Latin-1 header values are allowed.
+            resp = client.simulate_request(method, '/')
+            assert resp.headers
+        else:
+            if app._ASGI:
+                # NOTE(kgriffs): Unlike PEP-3333, the ASGI spec requires the
+                #   app to encode header names and values to a byte string. This
+                #   gives Falcon the opportunity to verify the character set
+                #   in the process and raise an error as appropriate.
+                error_type = ValueError
+            else:
+                # NOTE(kgriffs): The wsgiref validator that is integrated into
+                #   Falcon's testing framework will catch this. However, Falcon
+                #   itself does not do the check to avoid potential overhead
+                #   in a production deployment.
+                error_type = AssertionError
+
+            pattern = 'Bad header name' if method == 'POST' else 'Bad header value'
+
+            with pytest.raises(error_type, match=pattern):
+                client.simulate_request(method, '/')
 
     def test_response_set_and_get_header(self, client):
         resource = HeaderHelpersResource()
@@ -559,8 +702,7 @@ class TestHeaders(object):
 
     @pytest.mark.parametrize('vary,expected_value', [
         (['accept-encoding'], 'accept-encoding'),
-        ([u'accept-encoding', 'x-auth-token'], 'accept-encoding, x-auth-token'),
-        (('accept-encoding', u'x-auth-token'), 'accept-encoding, x-auth-token'),
+        (('accept-encoding', 'x-auth-token'), 'accept-encoding, x-auth-token'),
     ])
     def test_vary_header(self, client, vary, expected_value):
         resource = VaryHeaderResource(vary)
@@ -587,15 +729,15 @@ class TestHeaders(object):
         resource = XmlResource(content_type)
         self._check_header(client, resource, 'Content-Type', content_type)
 
-    def test_add_link_single(self, client):
+    def test_append_link_single(self, client):
         expected_value = '</things/2842>; rel=next'
 
         resource = LinkHeaderResource()
-        resource.add_link('/things/2842', 'next')
+        resource.append_link('/things/2842', 'next')
 
         self._check_link_header(client, resource, expected_value)
 
-    def test_add_link_multiple(self, client):
+    def test_append_link_multiple(self, client):
         expected_value = (
             '</things/2842>; rel=next, ' +
             '<http://%C3%A7runchy/bacon>; rel=contents, ' +
@@ -604,84 +746,84 @@ class TestHeaders(object):
             '<ab%C3%A7>; rel="https://example.com/too-%C3%A7runchy", ' +
             '</alt-thing>; rel="alternate http://example.com/%C3%A7runchy"')
 
-        uri = u'ab\u00e7' if compat.PY3 else 'ab\xc3\xa7'
+        uri = 'ab\u00e7'
 
         resource = LinkHeaderResource()
-        resource.add_link('/things/2842', 'next')
-        resource.add_link(u'http://\u00e7runchy/bacon', 'contents')
-        resource.add_link(uri, 'http://example.com/ext-type')
-        resource.add_link(uri, u'http://example.com/\u00e7runchy')
-        resource.add_link(uri, u'https://example.com/too-\u00e7runchy')
-        resource.add_link('/alt-thing',
-                          u'alternate http://example.com/\u00e7runchy')
+        resource.append_link('/things/2842', 'next')
+        resource.append_link('http://\u00e7runchy/bacon', 'contents')
+        resource.append_link(uri, 'http://example.com/ext-type')
+        resource.append_link(uri, 'http://example.com/\u00e7runchy')
+        resource.append_link(uri, 'https://example.com/too-\u00e7runchy')
+        resource.append_link('/alt-thing',
+                             'alternate http://example.com/\u00e7runchy')
 
         self._check_link_header(client, resource, expected_value)
 
-    def test_add_link_with_title(self, client):
+    def test_append_link_with_title(self, client):
         expected_value = ('</related/thing>; rel=item; '
                           'title="A related thing"')
 
         resource = LinkHeaderResource()
-        resource.add_link('/related/thing', 'item',
-                          title='A related thing')
+        resource.append_link('/related/thing', 'item',
+                             title='A related thing')
 
         self._check_link_header(client, resource, expected_value)
 
-    def test_add_link_with_title_star(self, client):
+    def test_append_link_with_title_star(self, client):
         expected_value = ('</related/thing>; rel=item; '
                           "title*=UTF-8''A%20related%20thing, "
                           '</%C3%A7runchy/thing>; rel=item; '
                           "title*=UTF-8'en'A%20%C3%A7runchy%20thing")
 
         resource = LinkHeaderResource()
-        resource.add_link('/related/thing', 'item',
-                          title_star=('', 'A related thing'))
+        resource.append_link('/related/thing', 'item',
+                             title_star=('', 'A related thing'))
 
-        resource.add_link(u'/\u00e7runchy/thing', 'item',
-                          title_star=('en', u'A \u00e7runchy thing'))
+        resource.append_link('/\u00e7runchy/thing', 'item',
+                             title_star=('en', 'A \u00e7runchy thing'))
 
         self._check_link_header(client, resource, expected_value)
 
-    def test_add_link_with_anchor(self, client):
+    def test_append_link_with_anchor(self, client):
         expected_value = ('</related/thing>; rel=item; '
                           'anchor="/some%20thing/or-other"')
 
         resource = LinkHeaderResource()
-        resource.add_link('/related/thing', 'item',
-                          anchor='/some thing/or-other')
+        resource.append_link('/related/thing', 'item',
+                             anchor='/some thing/or-other')
 
         self._check_link_header(client, resource, expected_value)
 
-    def test_add_link_with_hreflang(self, client):
+    def test_append_link_with_hreflang(self, client):
         expected_value = ('</related/thing>; rel=about; '
                           'hreflang=en')
 
         resource = LinkHeaderResource()
-        resource.add_link('/related/thing', 'about', hreflang='en')
+        resource.append_link('/related/thing', 'about', hreflang='en')
 
         self._check_link_header(client, resource, expected_value)
 
-    def test_add_link_with_hreflang_multi(self, client):
+    def test_append_link_with_hreflang_multi(self, client):
         expected_value = ('</related/thing>; rel=about; '
                           'hreflang=en-GB; hreflang=de')
 
         resource = LinkHeaderResource()
-        resource.add_link('/related/thing', 'about',
-                          hreflang=('en-GB', 'de'))
+        resource.append_link('/related/thing', 'about',
+                             hreflang=('en-GB', 'de'))
 
         self._check_link_header(client, resource, expected_value)
 
-    def test_add_link_with_type_hint(self, client):
+    def test_append_link_with_type_hint(self, client):
         expected_value = ('</related/thing>; rel=alternate; '
                           'type="video/mp4; codecs=avc1.640028"')
 
         resource = LinkHeaderResource()
-        resource.add_link('/related/thing', 'alternate',
-                          type_hint='video/mp4; codecs=avc1.640028')
+        resource.append_link('/related/thing', 'alternate',
+                             type_hint='video/mp4; codecs=avc1.640028')
 
         self._check_link_header(client, resource, expected_value)
 
-    def test_add_link_complex(self, client):
+    def test_append_link_complex(self, client):
         expected_value = ('</related/thing>; rel=alternate; '
                           'title="A related thing"; '
                           "title*=UTF-8'en'A%20%C3%A7runchy%20thing; "
@@ -689,20 +831,87 @@ class TestHeaders(object):
                           'hreflang=en-GB; hreflang=de')
 
         resource = LinkHeaderResource()
-        resource.add_link('/related/thing', 'alternate',
-                          title='A related thing',
-                          hreflang=('en-GB', 'de'),
-                          type_hint='application/json',
-                          title_star=('en', u'A \u00e7runchy thing'))
+        resource.append_link('/related/thing', 'alternate',
+                             title='A related thing',
+                             hreflang=('en-GB', 'de'),
+                             type_hint='application/json',
+                             title_star=('en', 'A \u00e7runchy thing'))
 
         self._check_link_header(client, resource, expected_value)
 
+    @pytest.mark.parametrize('crossorigin,expected_value', [
+        (None, '</related/thing>; rel=alternate'),
+        ('anonymous', '</related/thing>; rel=alternate; crossorigin'),
+        ('Anonymous', '</related/thing>; rel=alternate; crossorigin'),
+        ('AnOnYmOUs', '</related/thing>; rel=alternate; crossorigin'),
+        (
+            'Use-Credentials',
+            '</related/thing>; rel=alternate; crossorigin="use-credentials"',
+        ),
+        (
+            'use-credentials',
+            '</related/thing>; rel=alternate; crossorigin="use-credentials"',
+        ),
+    ])
+    def test_append_link_crossorigin(self, client, crossorigin, expected_value):
+        resource = LinkHeaderResource()
+        resource.append_link('/related/thing', 'alternate',
+                             crossorigin=crossorigin)
+
+        self._check_link_header(client, resource, expected_value)
+
+    @pytest.mark.parametrize('crossorigin', [
+        '*',
+        'Allow-all',
+        'Lax',
+        'MUST-REVALIDATE',
+        'Strict',
+        'deny',
+    ])
+    def test_append_link_invalid_crossorigin_value(self, crossorigin):
+        resp = falcon.Response()
+
+        with pytest.raises(ValueError):
+            resp.append_link('/related/resource', 'next', crossorigin=crossorigin)
+
     def test_content_length_options(self, client):
         result = client.simulate_options()
 
-        content_length = '0'
+        content_length = str(len(falcon.HTTPNotFound().to_json()))
         assert result.headers['Content-Length'] == content_length
 
+    def test_set_headers_with_custom_class(self, client):
+        client.app.add_route('/', CustomHeadersResource())
+
+        result = client.simulate_get('/')
+
+        assert 'test-header' in result.headers
+        assert result.headers['test-header'] == 'test-value'
+
+    def test_headers_with_custom_class_not_callable(self, client):
+        client.app.add_route('/', CustomHeadersResource())
+
+        result = client.simulate_post('/')
+
+        assert 'test-header' not in result.headers
+
+    def test_request_multiple_header(self, client):
+        resource = HeaderHelpersResource()
+        client.app.add_route('/', resource)
+
+        client.simulate_request(headers=[
+            # Singletone header; last one wins
+            ('Content-Type', 'text/plain'),
+            ('Content-Type', 'image/jpeg'),
+
+            # Should be concatenated
+            ('X-Thing', '1'),
+            ('X-Thing', '2'),
+        ])
+
+        assert resource.req.content_type == 'image/jpeg'
+        assert resource.req.get_header('X-Thing') == '1,2'
+
     # ----------------------------------------------------------------------
     # Helpers
     # ----------------------------------------------------------------------
