@@ -1,19 +1,24 @@
 # -*- coding: utf-8
 
 import datetime
+import json
+import wsgiref.validate
 import xml.etree.ElementTree as et  # noqa: I202
 
 import pytest
 import yaml
 
 import falcon
+from falcon.http_error import NoRepresentation, OptionalRepresentation
 import falcon.testing as testing
-from falcon.util import json
+from falcon.util.deprecation import DeprecatedWarning
+
+from _util import create_app  # NOQA
 
 
 @pytest.fixture
-def client():
-    app = falcon.API()
+def client(asgi):
+    app = create_app(asgi)
 
     resource = FaultyResource()
     app.add_route('/fail', resource)
@@ -29,19 +34,19 @@ class FaultyResource:
         description = req.get_header('X-Error-Description')
         code = 10042
 
-        raise falcon.HTTPError(status, title, description, code=code)
+        raise falcon.HTTPError(status, title=title, description=description, code=code)
 
     def on_post(self, req, resp):
         raise falcon.HTTPForbidden(
-            'Request denied',
-            'You do not have write permissions for this queue.',
+            title='Request denied',
+            description='You do not have write permissions for this queue.',
             href='http://example.com/api/rbac')
 
     def on_put(self, req, resp):
         raise falcon.HTTPError(
             falcon.HTTP_792,
-            'Internet crashed',
-            'Catastrophic weather event due to climate change.',
+            title='Internet crashed',
+            description='Catastrophic weather event due to climate change.',
             href='http://example.com/api/climate',
             href_text='Drill baby drill!',
             code=8733224)
@@ -50,7 +55,7 @@ class FaultyResource:
         raise falcon.HTTPError(falcon.HTTP_400)
 
 
-class UnicodeFaultyResource(object):
+class UnicodeFaultyResource:
 
     def __init__(self):
         self.called = False
@@ -59,10 +64,10 @@ class UnicodeFaultyResource(object):
         self.called = True
         raise falcon.HTTPError(
             falcon.HTTP_792,
-            u'Internet \xe7rashed!',
-            u'\xc7atastrophic weather event',
-            href=u'http://example.com/api/\xe7limate',
-            href_text=u'Drill b\xe1by drill!')
+            title='Internet \xe7rashed!',
+            description='\xc7atastrophic weather event',
+            href='http://example.com/api/\xe7limate',
+            href_text='Drill b\xe1by drill!')
 
 
 class MiscErrorsResource:
@@ -73,27 +78,27 @@ class MiscErrorsResource:
 
     def on_get(self, req, resp):
         if self.needs_title:
-            raise self._exception('Excuse Us', 'Something went boink!')
+            raise self._exception(title='Excuse Us', description='Something went boink!')
         else:
-            raise self._exception('Something went boink!')
+            raise self._exception(title='Something went boink!')
 
 
 class UnauthorizedResource:
 
     def on_get(self, req, resp):
-        raise falcon.HTTPUnauthorized('Authentication Required',
-                                      'Missing or invalid authorization.',
-                                      ['Basic realm="simple"'])
+        raise falcon.HTTPUnauthorized(title='Authentication Required',
+                                      description='Missing or invalid authorization.',
+                                      challenges=['Basic realm="simple"'])
 
     def on_post(self, req, resp):
-        raise falcon.HTTPUnauthorized('Authentication Required',
-                                      'Missing or invalid authorization.',
-                                      ['Newauth realm="apps"',
-                                       'Basic realm="simple"'])
+        raise falcon.HTTPUnauthorized(title='Authentication Required',
+                                      description='Missing or invalid authorization.',
+                                      challenges=['Newauth realm="apps"', 'Basic realm="simple"'])
 
     def on_put(self, req, resp):
-        raise falcon.HTTPUnauthorized('Authentication Required',
-                                      'Missing or invalid authorization.', [])
+        raise falcon.HTTPUnauthorized(title='Authentication Required',
+                                      description='Missing or invalid authorization.',
+                                      challenges=[])
 
 
 class NotFoundResource:
@@ -153,14 +158,14 @@ class MethodNotAllowedResourceWithBody:
 class LengthRequiredResource:
 
     def on_get(self, req, resp):
-        raise falcon.HTTPLengthRequired('title', 'description')
+        raise falcon.HTTPLengthRequired(title='title', description='description')
 
 
 class RequestEntityTooLongResource:
 
     def on_get(self, req, resp):
-        raise falcon.HTTPPayloadTooLarge('Request Rejected',
-                                         'Request Body Too Large')
+        raise falcon.HTTPPayloadTooLarge(title='Request Rejected',
+                                         description='Request Body Too Large')
 
 
 class TemporaryRequestEntityTooLongResource:
@@ -169,8 +174,8 @@ class TemporaryRequestEntityTooLongResource:
         self.retry_after = retry_after
 
     def on_get(self, req, resp):
-        raise falcon.HTTPPayloadTooLarge('Request Rejected',
-                                         'Request Body Too Large',
+        raise falcon.HTTPPayloadTooLarge(title='Request Rejected',
+                                         description='Request Body Too Large',
                                          retry_after=self.retry_after)
 
 
@@ -182,8 +187,8 @@ class UriTooLongResource:
         self.code = code
 
     def on_get(self, req, resp):
-        raise falcon.HTTPUriTooLong(self.title,
-                                    self.description,
+        raise falcon.HTTPUriTooLong(title=self.title,
+                                    description=self.description,
                                     code=self.code)
 
 
@@ -199,8 +204,8 @@ class TooManyRequestsResource:
         self.retry_after = retry_after
 
     def on_get(self, req, resp):
-        raise falcon.HTTPTooManyRequests('Too many requests',
-                                         '1 per minute',
+        raise falcon.HTTPTooManyRequests(title='Too many requests',
+                                         description='1 per minute',
                                          retry_after=self.retry_after)
 
 
@@ -210,8 +215,8 @@ class ServiceUnavailableResource:
         self.retry_after = retry_after
 
     def on_get(self, req, resp):
-        raise falcon.HTTPServiceUnavailable('Oops',
-                                            'Stand by...',
+        raise falcon.HTTPServiceUnavailable(title='Oops',
+                                            description='Stand by...',
                                             retry_after=self.retry_after)
 
 
@@ -243,7 +248,7 @@ class MissingParamResource:
         raise falcon.HTTPMissingParam('id', code='P1003')
 
 
-class TestHTTPError(object):
+class TestHTTPError:
 
     def _misc_test(self, client, exception, status, needs_title=True):
         client.app.add_route('/misc', MiscErrorsResource(exception, needs_title))
@@ -283,6 +288,10 @@ class TestHTTPError(object):
         assert response.status == headers['X-Error-Status']
         assert response.json == expected_body
 
+    def test_has_representation(self):
+        with pytest.warns(DeprecatedWarning, match='has_representation is deprecated'):
+            assert falcon.HTTPError(falcon.HTTP_701).has_representation is True
+
     def test_no_description_json(self, client):
         response = client.simulate_patch('/fail')
         assert response.status == falcon.HTTP_400
@@ -334,19 +343,14 @@ class TestHTTPError(object):
         }
 
         def _my_serializer(req, resp, exception):
-            representation = None
-
-            preferred = req.client_prefers(('application/x-yaml',
-                                            'application/json'))
+            preferred = req.client_prefers((falcon.MEDIA_YAML, falcon.MEDIA_JSON))
 
             if preferred is not None:
-                if preferred == 'application/json':
-                    representation = exception.to_json()
+                if preferred == falcon.MEDIA_JSON:
+                    resp.data = exception.to_json()
                 else:
-                    representation = yaml.dump(exception.to_dict(),
-                                               encoding=None)
+                    resp.text = yaml.dump(exception.to_dict(), encoding=None)
 
-                resp.body = representation
                 resp.content_type = preferred
 
         def _check(media_type, deserializer):
@@ -358,8 +362,8 @@ class TestHTTPError(object):
             actual_doc = deserializer(response.content.decode('utf-8'))
             assert expected_doc == actual_doc
 
-        _check('application/x-yaml', yaml.load)
-        _check('application/json', json.loads)
+        _check(falcon.MEDIA_YAML, yaml.safe_load)
+        _check(falcon.MEDIA_JSON, json.loads)
 
     @pytest.mark.parametrize('method,path,status', [
         ('GET', '/404', 404),
@@ -379,14 +383,25 @@ class TestHTTPError(object):
         client.app.add_route('/notfound', NotFoundResourceWithBody())
         client.app.set_error_serializer(_simple_serializer)
 
-        resp = client.simulate_request(path=path, method=method)
+        def s():
+            return client.simulate_request(path=path, method=method)
+
+        if method not in falcon.COMBINED_METHODS:
+            if not client.app._ASGI:
+                with pytest.warns(wsgiref.validate.WSGIWarning):
+                    resp = s()
+            else:
+                resp = s()
+        else:
+            resp = s()
+
         assert resp.json['title']
         assert resp.json['status'] == status
 
     def test_custom_serializer_no_representation(self, client):
         def _chatty_serializer(req, resp, exception):
             resp.content_type = falcon.MEDIA_TEXT
-            resp.body = b'You might think this error should not have a body'
+            resp.text = b'You might think this error should not have a body'
 
         client.app.add_route('/416', RangeNotSatisfiableResource())
         client.app.set_error_serializer(_chatty_serializer)
@@ -486,10 +501,10 @@ class TestHTTPError(object):
         unicode_resource = UnicodeFaultyResource()
 
         expected_body = {
-            'title': u'Internet \xe7rashed!',
-            'description': u'\xc7atastrophic weather event',
+            'title': 'Internet \xe7rashed!',
+            'description': '\xc7atastrophic weather event',
             'link': {
-                'text': u'Drill b\xe1by drill!',
+                'text': 'Drill b\xe1by drill!',
                 'href': 'http://example.com/api/%C3%A7limate',
                 'rel': 'help',
             },
@@ -505,18 +520,18 @@ class TestHTTPError(object):
     def test_unicode_xml(self, client):
         unicode_resource = UnicodeFaultyResource()
 
-        expected_body = (u'<?xml version="1.0" encoding="UTF-8"?>' +
-                         u'<error>' +
-                         u'<title>Internet çrashed!</title>' +
-                         u'<description>' +
-                         u'Çatastrophic weather event' +
-                         u'</description>' +
-                         u'<link>' +
-                         u'<text>Drill báby drill!</text>' +
-                         u'<href>http://example.com/api/%C3%A7limate</href>' +
-                         u'<rel>help</rel>' +
-                         u'</link>' +
-                         u'</error>')
+        expected_body = ('<?xml version="1.0" encoding="UTF-8"?>' +
+                         '<error>' +
+                         '<title>Internet çrashed!</title>' +
+                         '<description>' +
+                         'Çatastrophic weather event' +
+                         '</description>' +
+                         '<link>' +
+                         '<text>Drill báby drill!</text>' +
+                         '<href>http://example.com/api/%C3%A7limate</href>' +
+                         '<rel>help</rel>' +
+                         '</link>' +
+                         '</error>')
 
         client.app.add_route('/unicode', unicode_resource)
         response = client.simulate_request(
@@ -550,7 +565,8 @@ class TestHTTPError(object):
         response = client.simulate_request(path='/404')
 
         assert response.status == falcon.HTTP_404
-        assert not response.content
+        assert response.json == falcon.HTTPNotFound().to_dict()
+        assert response.json == {'title': falcon.HTTP_NOT_FOUND}
 
     def test_404_with_body(self, client):
         client.app.add_route('/404', NotFoundResourceWithBody())
@@ -559,8 +575,8 @@ class TestHTTPError(object):
         assert response.status == falcon.HTTP_404
         assert response.content
         expected_body = {
-            u'title': u'404 Not Found',
-            u'description': u'Not Found'
+            'title': '404 Not Found',
+            'description': 'Not Found'
         }
         assert response.json == expected_body
 
@@ -569,7 +585,8 @@ class TestHTTPError(object):
 
         response = client.simulate_request(path='/405')
         assert response.status == falcon.HTTP_405
-        assert not response.content
+        assert response.content == falcon.HTTPMethodNotAllowed(['PUT']).to_json()
+        assert response.json == {'title': falcon.HTTP_METHOD_NOT_ALLOWED}
         assert response.headers['allow'] == 'PUT'
 
     def test_405_without_body_with_extra_headers(self, client):
@@ -577,18 +594,18 @@ class TestHTTPError(object):
 
         response = client.simulate_request(path='/405')
         assert response.status == falcon.HTTP_405
-        assert not response.content
+        assert response.content == falcon.HTTPMethodNotAllowed([]).to_json()
         assert response.headers['allow'] == 'PUT'
         assert response.headers['x-ping'] == 'pong'
 
     def test_405_without_body_with_extra_headers_double_check(self, client):
         client.app.add_route(
-            '/405/', MethodNotAllowedResourceWithHeadersWithAccept()
+            '/405', MethodNotAllowedResourceWithHeadersWithAccept()
         )
 
         response = client.simulate_request(path='/405')
         assert response.status == falcon.HTTP_405
-        assert not response.content
+        assert response.json == falcon.HTTPMethodNotAllowed([]).to_dict()
         assert response.headers['allow'] == 'PUT'
         assert response.headers['allow'] != 'GET,PUT'
         assert response.headers['allow'] != 'GET'
@@ -601,8 +618,8 @@ class TestHTTPError(object):
         assert response.status == falcon.HTTP_405
         assert response.content
         expected_body = {
-            u'title': u'405 Method Not Allowed',
-            u'description': u'Not Allowed'
+            'title': '405 Method Not Allowed',
+            'description': 'Not Allowed'
         }
         assert response.json == expected_body
         assert response.headers['allow'] == 'PUT'
@@ -612,7 +629,8 @@ class TestHTTPError(object):
         response = client.simulate_request(path='/410')
 
         assert response.status == falcon.HTTP_410
-        assert not response.content
+        assert response.content == falcon.HTTPGone().to_json()
+        assert response.json == {'title': '410 Gone'}
 
     def test_410_with_body(self, client):
         client.app.add_route('/410', GoneResourceWithBody())
@@ -621,8 +639,8 @@ class TestHTTPError(object):
         assert response.status == falcon.HTTP_410
         assert response.content
         expected_body = {
-            u'title': u'410 Gone',
-            u'description': u'Gone with the wind'
+            'title': '410 Gone',
+            'description': 'Gone with the wind'
         }
         assert response.json == expected_body
 
@@ -696,15 +714,20 @@ class TestHTTPError(object):
         parsed_body = json.loads(response.content.decode())
         assert parsed_body['code'] == code
 
-    def test_416(self, client):
-        client.app = falcon.API()
+    def test_416(self, client, asgi):
+        client.app = create_app(asgi)
         client.app.add_route('/416', RangeNotSatisfiableResource())
         response = client.simulate_request(path='/416', headers={'accept': 'text/xml'})
 
         assert response.status == falcon.HTTP_416
-        assert not response.content
+        assert response.content == falcon.HTTPRangeNotSatisfiable(123456).to_xml()
+        exp = (
+            b'<?xml version="1.0" encoding="UTF-8"?><error>'
+            b'<title>416 Range Not Satisfiable</title></error>'
+        )
+        assert response.content == exp
         assert response.headers['content-range'] == 'bytes */123456'
-        assert response.headers['content-length'] == '0'
+        assert response.headers['content-length'] == str(len(response.content))
 
     def test_429_no_retry_after(self, client):
         client.app.add_route('/429', TooManyRequestsResource())
@@ -742,8 +765,8 @@ class TestHTTPError(object):
         response = client.simulate_request(path='/503')
 
         expected_body = {
-            u'title': u'Oops',
-            u'description': u'Stand by...',
+            'title': 'Oops',
+            'description': 'Stand by...',
         }
 
         assert response.status == falcon.HTTP_503
@@ -756,8 +779,8 @@ class TestHTTPError(object):
         response = client.simulate_request(path='/503')
 
         expected_body = {
-            u'title': u'Oops',
-            u'description': u'Stand by...',
+            'title': 'Oops',
+            'description': 'Stand by...',
         }
 
         assert response.status == falcon.HTTP_503
@@ -768,13 +791,13 @@ class TestHTTPError(object):
         client.app.add_route('/400', InvalidHeaderResource())
         response = client.simulate_request(path='/400')
 
-        expected_desc = (u'The value provided for the X-Auth-Token '
-                         u'header is invalid. Please provide a valid token.')
+        expected_desc = ('The value provided for the "X-Auth-Token" '
+                         'header is invalid. Please provide a valid token.')
 
         expected_body = {
-            u'title': u'Invalid header value',
-            u'description': expected_desc,
-            u'code': u'A1001',
+            'title': 'Invalid header value',
+            'description': expected_desc,
+            'code': 'A1001',
         }
 
         assert response.status == falcon.HTTP_400
@@ -785,8 +808,8 @@ class TestHTTPError(object):
         response = client.simulate_request(path='/400')
 
         expected_body = {
-            u'title': u'Missing header value',
-            u'description': u'The X-Auth-Token header is required.',
+            'title': 'Missing header value',
+            'description': 'The "X-Auth-Token" header is required.',
         }
 
         assert response.status == falcon.HTTP_400
@@ -796,12 +819,12 @@ class TestHTTPError(object):
         client.app.add_route('/400', InvalidParamResource())
         response = client.simulate_request(path='/400')
 
-        expected_desc = (u'The "id" parameter is invalid. The '
-                         u'value must be a hex-encoded UUID.')
+        expected_desc = ('The "id" parameter is invalid. The '
+                         'value must be a hex-encoded UUID.')
         expected_body = {
-            u'title': u'Invalid parameter',
-            u'description': expected_desc,
-            u'code': u'P1002',
+            'title': 'Invalid parameter',
+            'description': expected_desc,
+            'code': 'P1002',
         }
 
         assert response.status == falcon.HTTP_400
@@ -812,9 +835,9 @@ class TestHTTPError(object):
         response = client.simulate_request(path='/400')
 
         expected_body = {
-            u'title': u'Missing parameter',
-            u'description': u'The "id" parameter is required.',
-            u'code': u'P1003',
+            'title': 'Missing parameter',
+            'description': 'The "id" parameter is required.',
+            'code': 'P1003',
         }
 
         assert response.status == falcon.HTTP_400
@@ -843,3 +866,76 @@ class TestHTTPError(object):
 
         assert response.status == headers['X-Error-Status']
         assert response.json['title'] == headers['X-Error-Status']
+
+    def test_to_json_dumps(self):
+        e = falcon.HTTPError(status=falcon.HTTP_418, title='foo', description='bar')
+        assert e.to_json() == b'{"title": "foo", "description": "bar"}'
+
+        class Handler:
+            def serialize(self, obj, type):
+                assert type == falcon.MEDIA_JSON
+                return b'{"a": "b"}'
+
+        assert e.to_json(Handler()) == b'{"a": "b"}'
+
+    def test_serialize_error_uses_media_handler(self, client):
+        client.app.add_route('/path', NotFoundResource())
+        h = client.app.resp_options.media_handlers[falcon.MEDIA_JSON]
+        h._dumps = lambda x: json.dumps(x).upper()
+        response = client.simulate_request(path='/path')
+
+        assert response.status == falcon.HTTP_404
+        assert response.json == {'TITLE': falcon.HTTP_NOT_FOUND.upper()}
+
+    def test_serialize_no_json_media_handler(self, client):
+        client.app.add_route('/path', NotFoundResource())
+        for h in list(client.app.resp_options.media_handlers):
+            if 'json' in h.casefold():
+                client.app.resp_options.media_handlers.pop(h)
+        response = client.simulate_request(path='/path')
+
+        assert response.status == falcon.HTTP_404
+        assert response.json == {'title': falcon.HTTP_NOT_FOUND}
+
+    def test_MediaMalformedError(self):
+        err = falcon.MediaMalformedError('foo-media')
+        assert err.description == 'Could not parse foo-media body'
+
+        err.__cause__ = ValueError('some error')
+        assert err.description == 'Could not parse foo-media body - some error'
+
+
+def test_kw_only():
+    # only deprecated for now
+    # with pytest.raises(TypeError, match='positional argument'):
+    #     falcon.HTTPError(falcon.HTTP_BAD_REQUEST, 'foo', 'bar')
+    with pytest.warns(DeprecatedWarning, match='positional args are deprecated'):
+        falcon.HTTPError(falcon.HTTP_BAD_REQUEST, 'foo', 'bar')
+
+
+def test_NoRepresentation():
+    with pytest.warns(
+        DeprecatedWarning,
+        match='has_representation is deprecated.*The class NoRepresentation'
+    ):
+        assert NoRepresentation().has_representation is False
+
+
+class TestOptionalRepresentation:
+    def test_OptionalRepresentation_false(self):
+        with pytest.warns(
+            DeprecatedWarning,
+            match='has_representation is deprecated.*The class OptionalRepresentation'
+        ):
+            or_ = OptionalRepresentation()
+            or_.description = None
+            assert or_.has_representation is False
+
+    def test_OptionalRepresentation_true(self):
+        with pytest.warns(
+            DeprecatedWarning,
+            match='has_representation is deprecated.*The class OptionalRepresentation'
+        ):
+            or_ = OptionalRepresentation()
+            or_.description = 'foo'
+            assert or_.has_representation is True
