@@ -8,6 +8,8 @@ import falcon
 from falcon.errors import HTTPInvalidParam
 import falcon.testing as testing
 
+from _util import create_app  # NOQA
+
 
 class Resource(testing.SimpleTestResource):
 
@@ -43,9 +45,11 @@ def resource():
 
 
 @pytest.fixture
-def client():
-    app = falcon.API()
-    app.req_options.auto_parse_form_urlencoded = True
+def client(asgi):
+    app = create_app(asgi)
+    if not asgi:
+        app.req_options.auto_parse_form_urlencoded = True
+
     return testing.TestClient(app)
 
 
@@ -54,6 +58,12 @@ def simulate_request_get_query_params(client, path, query_string, **kwargs):
 
 
 def simulate_request_post_query_params(client, path, query_string, **kwargs):
+    if client.app._ASGI:
+        pytest.skip(
+            'The ASGI implementation does not support '
+            'RequestOptions.auto_parse_form_urlencoded'
+        )
+
     headers = kwargs.setdefault('headers', {})
     headers['Content-Type'] = 'application/x-www-form-urlencoded'
     if 'method' not in kwargs:
@@ -72,7 +82,7 @@ def simulate_request(request):
     return request.param
 
 
-class TestQueryParams(object):
+class TestQueryParams:
     def test_none(self, simulate_request, client, resource):
         query_string = ''
         client.app.add_route('/', resource)  # TODO: DRY up this setup logic
@@ -140,10 +150,10 @@ class TestQueryParams(object):
 
         # NOTE(kgriffs): For lists, get_param will return one of the
         # elements, but which one it will choose is undefined.
-        assert req.get_param('id') in [u'23', u'42']
+        assert req.get_param('id') in ['23', '42']
 
         assert req.get_param_as_list('id', int) == [23, 42]
-        assert req.get_param('q') == u'\u8c46 \u74e3'
+        assert req.get_param('q') == '\u8c46 \u74e3'
 
     def test_option_auto_parse_qs_csv_simple_false(self, simulate_request, client, resource):
         client.app.add_route('/', resource)
@@ -154,9 +164,9 @@ class TestQueryParams(object):
 
         req = resource.captured_req
 
-        assert req.params['id'] == [u'23,42,,', u'2']
-        assert req.get_param('id') in [u'23,42,,', u'2']
-        assert req.get_param_as_list('id') == [u'23,42,,', u'2']
+        assert req.params['id'] == ['23,42,,', '2']
+        assert req.get_param('id') in ['23,42,,', '2']
+        assert req.get_param_as_list('id') == ['23,42,,', '2']
 
     def test_option_auto_parse_qs_csv_simple_true(self, simulate_request, client, resource):
         client.app.add_route('/', resource)
@@ -168,10 +178,64 @@ class TestQueryParams(object):
 
         req = resource.captured_req
 
-        assert req.params['id'] == [u'23', u'42', u'2']
-        assert req.get_param('id') in [u'23', u'42', u'2']
+        assert req.params['id'] == ['23', '42', '2']
+        assert req.get_param('id') in ['23', '42', '2']
         assert req.get_param_as_list('id', int) == [23, 42, 2]
 
+    def test_option_auto_parse_qs_csv_multiple_fields_false(
+        self,
+        simulate_request,
+        client,
+        resource
+    ):
+        client.app.add_route('/', resource)
+
+        query_string = 't=1,2&t=3,4'
+        simulate_request(client=client, path='/', query_string=query_string)
+
+        req = resource.captured_req
+
+        assert req.params['t'] == ['1,2', '3,4']
+        assert req.get_param('t') in ['1,2', '3,4']
+        assert req.get_param_as_list('t') == ['1,2', '3,4']
+
+    @pytest.mark.parametrize(
+        'qs, keep_blank, expected',
+        [
+            ('t=1&t=3,4', False, ['1', '3', '4']),
+            ('t=1&t=2&t=3,4', False, ['1', '2', '3', '4']),
+            ('t=1,2&t=3,4', False, ['1', '2', '3', '4']),
+            ('t=1,,2&t=3,4', False, ['1', '2', '3', '4']),
+            ('t=1,,2&t=3,4', True, ['1', '', '2', '3', '4']),
+            ('t=1,2&t=3,4,,5', False, ['1', '2', '3', '4', '5']),
+            ('t=1&t=,1,4,,5', False, ['1', '1', '4', '5']),
+            ('t=1&t=,1,4,,5', True, ['1', '', '1', '4', '', '5']),
+            ('t=1&t=,1,4,,5&t=a,b,c', True, ['1', '', '1', '4', '', '5', 'a', 'b', 'c']),
+        ]
+    )
+    def test_option_auto_parse_qs_csv_multiple_fields_true(
+        self,
+        simulate_request,
+        client,
+        resource,
+
+        qs,
+        keep_blank,
+        expected,
+    ):
+        client.app.add_route('/', resource)
+        client.app.req_options.auto_parse_qs_csv = True
+        client.app.req_options.keep_blank_qs_values = keep_blank
+
+        query_string = qs
+        simulate_request(client=client, path='/', query_string=query_string)
+
+        req = resource.captured_req
+
+        assert req.params['t'] == expected
+        assert req.get_param('t') in expected
+        assert req.get_param_as_list('t') == expected
+
     def test_option_auto_parse_qs_csv_complex_false(self, simulate_request, client, resource):
         client.app.add_route('/', resource)
         client.app.req_options.auto_parse_qs_csv = False
@@ -190,17 +254,17 @@ class TestQueryParams(object):
         req = resource.captured_req
 
         assert req.get_param('colors') in 'red,green,blue'
-        assert req.get_param_as_list('colors') == [u'red,green,blue']
+        assert req.get_param_as_list('colors') == ['red,green,blue']
 
         assert req.get_param_as_list('limit') == ['1']
 
         assert req.get_param_as_list('empty1') is None
-        assert req.get_param_as_list('empty2') == [u',']
-        assert req.get_param_as_list('empty3') == [u',,']
+        assert req.get_param_as_list('empty2') == [',']
+        assert req.get_param_as_list('empty3') == [',,']
 
-        assert req.get_param_as_list('list-ish1') == [u'f,,x']
-        assert req.get_param_as_list('list-ish2') == [u',0']
-        assert req.get_param_as_list('list-ish3') == [u'a,,,b']
+        assert req.get_param_as_list('list-ish1') == ['f,,x']
+        assert req.get_param_as_list('list-ish2') == [',0']
+        assert req.get_param_as_list('list-ish3') == ['a,,,b']
 
         assert req.get_param('thing') == decoded_json
 
@@ -450,9 +514,25 @@ class TestQueryParams(object):
     def test_boolean(self, simulate_request, client, resource):
         client.app.add_route('/', resource)
         client.app.req_options.keep_blank_qs_values = False
-        query_string = ('echo=true&doit=false&bogus=bar&bogus2=foo&'
-                        't1=True&f1=False&t2=yes&f2=no&blank&one=1&zero=0&'
-                        'checkbox1=on&checkbox2=off')
+        query_string = '&'.join([
+            'echo=true',
+            'doit=false',
+            'bogus=bar',
+            'bogus2=foo',
+            't1=True',
+            'f1=False',
+            't2=yes',
+            'f2=no',
+            't3=y',
+            'f3=n',
+            't4=t',
+            'f4=f',
+            'blank',
+            'one=1',
+            'zero=0',
+            'checkbox1=on',
+            'checkbox2=off',
+        ])
         simulate_request(client=client, path='/', query_string=query_string)
 
         req = resource.captured_req
@@ -472,10 +552,10 @@ class TestQueryParams(object):
         assert req.get_param_as_bool('echo') is True
         assert req.get_param_as_bool('doit') is False
 
-        assert req.get_param_as_bool('t1') is True
-        assert req.get_param_as_bool('t2') is True
-        assert req.get_param_as_bool('f1') is False
-        assert req.get_param_as_bool('f2') is False
+        for i in range(1, 5):
+            assert req.get_param_as_bool('t' + str(i)) is True
+            assert req.get_param_as_bool('f' + str(i)) is False
+
         assert req.get_param_as_bool('one') is True
         assert req.get_param_as_bool('zero') is False
         assert req.get_param('blank') is None
@@ -815,6 +895,24 @@ class TestQueryParams(object):
         with pytest.raises(HTTPInvalidParam):
             req.get_param_as_json('payload')
 
+    def test_get_param_as_json_handler_json(self, client, resource):
+        client.app.add_route('/', resource)
+        payload_dict = {'foo': 'bar'}
+        query_string = 'payload={}'.format(json.dumps(payload_dict))
+        client.app.req_options.media_handlers[falcon.MEDIA_JSON]._loads = lambda x: {'x': 'y'}
+        client.simulate_get(path='/', query_string=query_string)
+        req = resource.captured_req
+        assert req.get_param_as_json('payload') == {'x': 'y'}
+
+    def test_get_param_as_json_no_handler_json(self, client, resource):
+        client.app.add_route('/', resource)
+        payload_dict = {'foo': 'bar'}
+        query_string = 'payload={}'.format(json.dumps(payload_dict))
+        client.app.req_options.media_handlers.pop(falcon.MEDIA_JSON)
+        client.simulate_get(path='/', query_string=query_string)
+        req = resource.captured_req
+        assert req.get_param_as_json('payload') == {'foo': 'bar'}
+
     def test_has_param(self, simulate_request, client, resource):
         client.app.add_route('/', resource)
         query_string = 'ant=1'
@@ -829,7 +927,7 @@ class TestQueryParams(object):
         assert not req.has_param(None)
 
 
-class TestPostQueryParams(object):
+class TestPostQueryParams:
     @pytest.mark.parametrize('http_method', ('POST', 'PUT', 'PATCH', 'DELETE', 'OPTIONS'))
     def test_http_methods_body_expected(self, client, resource, http_method):
         client.app.add_route('/', resource)
@@ -854,7 +952,7 @@ class TestPostQueryParams(object):
 
     def test_non_ascii(self, client, resource):
         client.app.add_route('/', resource)
-        value = u'\u8c46\u74e3'
+        value = '\u8c46\u74e3'
         query_string = b'q=' + value.encode('utf-8')
         simulate_request_post_query_params(client=client, path='/', query_string=query_string)
 
@@ -883,10 +981,20 @@ class TestPostQueryParams(object):
         req = resource.captured_req
         assert req.get_param('q') is None
 
+    def test_asgi_raises_error(self, resource):
+        app = create_app(asgi=True)
+        app.add_route('/', resource)
+        app.req_options.auto_parse_form_urlencoded = True
 
-class TestPostQueryParamsDefaultBehavior(object):
-    def test_dont_auto_parse_by_default(self):
-        app = falcon.API()
+        with pytest.raises(RuntimeError) as exc_info:
+            testing.simulate_get(app, '/')
+        assert 'RequestOptions.auto_parse_form_urlencoded' in exc_info.value.args[0]
+
+
+@pytest.mark.parametrize('asgi', [True, False])
+class TestPostQueryParamsDefaultBehavior:
+    def test_dont_auto_parse_by_default(self, asgi):
+        app = create_app(asgi)
         resource = testing.SimpleTestResource()
         app.add_route('/', resource)
 
