@@ -3,10 +3,12 @@ import pytest
 import falcon
 import falcon.testing as testing
 
+from _util import create_app  # NOQA
+
 
 @pytest.fixture
-def client():
-    app = falcon.API()
+def client(asgi):
+    app = create_app(asgi)
 
     resource = RedirectingResource()
     app.add_route('/', resource)
@@ -15,8 +17,8 @@ def client():
 
 
 @pytest.fixture
-def client_exercising_headers():
-    app = falcon.API()
+def client_exercising_headers(asgi):
+    app = create_app(asgi)
 
     resource = RedirectingResourceWithHeaders()
     app.add_route('/', resource)
@@ -24,7 +26,7 @@ def client_exercising_headers():
     return testing.TestClient(app)
 
 
-class RedirectingResource(object):
+class RedirectingResource:
     # NOTE(kgriffs): You wouldn't necessarily use these types of
     # http methods with these types of redirects; this is only
     # done to simplify testing.
@@ -45,7 +47,7 @@ class RedirectingResource(object):
         raise falcon.HTTPPermanentRedirect('/perm/redirect')
 
 
-class RedirectingResourceWithHeaders(object):
+class RedirectingResourceWithHeaders:
     # NOTE(kgriffs): You wouldn't necessarily use these types of
     # http methods with these types of redirects; this is only
     # done to simplify testing.
@@ -66,7 +68,7 @@ class RedirectingResourceWithHeaders(object):
         raise falcon.HTTPPermanentRedirect('/perm/redirect', headers={'foo': 'bar'})
 
 
-class TestRedirects(object):
+class TestRedirects:
     @pytest.mark.parametrize('method,expected_status,expected_location', [
         ('GET', falcon.HTTP_301, '/moved/perm'),
         ('POST', falcon.HTTP_302, '/found'),
