@@ -6,6 +6,7 @@ import pytest
 import falcon
 from falcon import request_helpers
 import falcon.request
+from falcon.stream import BoundedStream
 import falcon.testing as testing
 
 SIZE_1_KB = 1024
@@ -18,20 +19,20 @@ def resource():
 
 @pytest.fixture
 def client():
-    app = falcon.API()
+    app = falcon.App()
     return testing.TestClient(app)
 
 
-class TestRequestBody(object):
+class TestRequestBody:
     def _get_wrapped_stream(self, req):
         # Getting wrapped wsgi.input:
         stream = req.stream
-        if isinstance(stream, request_helpers.BoundedStream):
+        if isinstance(stream, BoundedStream):
             stream = stream.stream
         if isinstance(stream, InputWrapper):
             stream = stream.input
-        if isinstance(stream, io.BytesIO):
-            return stream
+
+        return stream
 
     def test_empty_body(self, client, resource):
         client.app.add_route('/', resource)
