@@ -6,16 +6,23 @@ import os
 import pytest
 
 import falcon
-from falcon.request import Request
-from falcon.response import Response
-from falcon.routing import StaticRoute
+from falcon.routing import StaticRoute, StaticRouteAsync
 import falcon.testing as testing
 
+import _util  # NOQA
 
-@pytest.fixture
-def client():
-    app = falcon.API()
-    return testing.TestClient(app)
+
+@pytest.fixture()
+def client(asgi):
+    app = _util.create_app(asgi=asgi)
+    client = testing.TestClient(app)
+    client.asgi = asgi
+    return client
+
+
+def create_sr(asgi, *args, **kwargs):
+    sr_type = StaticRouteAsync if asgi else StaticRoute
+    return sr_type(*args, **kwargs)
 
 
 @pytest.mark.parametrize('uri', [
@@ -76,26 +83,28 @@ def client():
     '/static/' + ('t' * StaticRoute._MAX_NON_PREFIXED_LEN) + 'x',
 
     # Invalid unicode character
-    #
-    # NOTE(kgriffs) create_environ() will take care of encoding this path
-    #   to UTF-8 under Python 2.
-    u'/static/\ufffdsomething',
+    '/static/\ufffdsomething',
 ])
-def test_bad_path(uri, monkeypatch):
-    monkeypatch.setattr(io, 'open', lambda path, mode: path)
+def test_bad_path(asgi, uri, monkeypatch):
+    monkeypatch.setattr(io, 'open', lambda path, mode: io.BytesIO())
 
-    sr = StaticRoute('/static', '/var/www/statics')
+    sr_type = StaticRouteAsync if asgi else StaticRoute
+    sr = sr_type('/static', '/var/www/statics')
 
-    req = Request(testing.create_environ(
+    req = _util.create_req(
+        asgi,
         host='test.com',
         path=uri,
-        app='statics'
-    ))
+        root_path='statics'
+    )
 
-    resp = Response()
+    resp = _util.create_resp(asgi)
 
     with pytest.raises(falcon.HTTPNotFound):
-        sr(req, resp)
+        if asgi:
+            falcon.async_to_sync(sr, req, resp)
+        else:
+            sr(req, resp)
 
 
 @pytest.mark.parametrize('prefix, directory', [
@@ -104,9 +113,9 @@ def test_bad_path(uri, monkeypatch):
     ('/static', 'statics'),
     ('/static', '../statics'),
 ])
-def test_invalid_args(prefix, directory, client):
+def test_invalid_args(client, prefix, directory):
     with pytest.raises(ValueError):
-        StaticRoute(prefix, directory)
+        create_sr(client.asgi, prefix, directory)
 
     with pytest.raises(ValueError):
         client.app.add_static_route(prefix, directory)
@@ -121,12 +130,18 @@ def test_invalid_args(prefix, directory, client):
 def test_invalid_args_fallback_filename(client, default):
     prefix, directory = '/static', '/var/www/statics'
     with pytest.raises(ValueError, match='fallback_filename'):
-        StaticRoute(prefix, directory, fallback_filename=default)
+        create_sr(client.asgi, prefix, directory, fallback_filename=default)
 
     with pytest.raises(ValueError, match='fallback_filename'):
         client.app.add_static_route(prefix, directory, fallback_filename=default)
 
 
+# NOTE(caselit) depending on the system configuration mime types can have alternative names
+_MIME_ALTERNATIVE = {
+    'application/zip': ('application/zip', 'application/x-zip-compressed')
+}
+
+
 @pytest.mark.parametrize('uri_prefix, uri_path, expected_path, mtype', [
     ('/static/', '/css/test.css', '/css/test.css', 'text/css'),
     ('/static', '/css/test.css', '/css/test.css', 'text/css'),
@@ -144,60 +159,69 @@ def test_invalid_args_fallback_filename(client, default):
     ('/some/download', '/foo/../bar/../report.zip', '/report.zip', 'application/zip'),
     ('/some/download', '/foo/bar/../../report.zip', '/report.zip', 'application/zip'),
 ])
-def test_good_path(uri_prefix, uri_path, expected_path, mtype, monkeypatch):
-    monkeypatch.setattr(io, 'open', lambda path, mode: path)
+def test_good_path(asgi, uri_prefix, uri_path, expected_path, mtype, monkeypatch):
+    monkeypatch.setattr(io, 'open', lambda path, mode: io.BytesIO(path.encode()))
 
-    sr = StaticRoute(uri_prefix, '/var/www/statics')
+    sr = create_sr(asgi, uri_prefix, '/var/www/statics')
 
     req_path = uri_prefix[:-1] if uri_prefix.endswith('/') else uri_prefix
     req_path += uri_path
 
-    req = Request(testing.create_environ(
+    req = _util.create_req(
+        asgi,
         host='test.com',
         path=req_path,
-        app='statics'
-    ))
+        root_path='statics'
+    )
 
-    resp = Response()
+    resp = _util.create_resp(asgi)
 
-    sr(req, resp)
+    if asgi:
+        async def run():
+            await sr(req, resp)
+            return await resp.stream.read()
 
-    assert resp.content_type == mtype
-    assert resp.stream == '/var/www/statics' + expected_path
+        body = falcon.async_to_sync(run)
+    else:
+        sr(req, resp)
+        body = resp.stream.read()
+
+    assert resp.content_type in _MIME_ALTERNATIVE.get(mtype, (mtype, ))
+    assert body.decode() == os.path.normpath('/var/www/statics' + expected_path)
 
 
 def test_lifo(client, monkeypatch):
-    monkeypatch.setattr(io, 'open', lambda path, mode: [path.encode('utf-8')])
+    monkeypatch.setattr(io, 'open', lambda path, mode: io.BytesIO(path.encode()))
 
     client.app.add_static_route('/downloads', '/opt/somesite/downloads')
     client.app.add_static_route('/downloads/archive', '/opt/somesite/x')
 
     response = client.simulate_request(path='/downloads/thing.zip')
     assert response.status == falcon.HTTP_200
-    assert response.text == '/opt/somesite/downloads/thing.zip'
+    assert response.text == os.path.normpath('/opt/somesite/downloads/thing.zip')
 
     response = client.simulate_request(path='/downloads/archive/thingtoo.zip')
     assert response.status == falcon.HTTP_200
-    assert response.text == '/opt/somesite/x/thingtoo.zip'
+    assert response.text == os.path.normpath('/opt/somesite/x/thingtoo.zip')
 
 
 def test_lifo_negative(client, monkeypatch):
-    monkeypatch.setattr(io, 'open', lambda path, mode: [path.encode('utf-8')])
+    monkeypatch.setattr(io, 'open', lambda path, mode: io.BytesIO(path.encode()))
 
     client.app.add_static_route('/downloads/archive', '/opt/somesite/x')
     client.app.add_static_route('/downloads', '/opt/somesite/downloads')
 
     response = client.simulate_request(path='/downloads/thing.zip')
     assert response.status == falcon.HTTP_200
-    assert response.text == '/opt/somesite/downloads/thing.zip'
+    assert response.text == os.path.normpath('/opt/somesite/downloads/thing.zip')
 
     response = client.simulate_request(path='/downloads/archive/thingtoo.zip')
     assert response.status == falcon.HTTP_200
-    assert response.text == '/opt/somesite/downloads/archive/thingtoo.zip'
+    assert response.text == os.path.normpath('/opt/somesite/downloads/archive/thingtoo.zip')
 
 
 def test_downloadable(client, monkeypatch):
-    monkeypatch.setattr(io, 'open', lambda path, mode: [path.encode('utf-8')])
+    monkeypatch.setattr(io, 'open', lambda path, mode: io.BytesIO(path.encode()))
 
     client.app.add_static_route('/downloads', '/opt/somesite/downloads', downloadable=True)
     client.app.add_static_route('/assets/', '/opt/somesite/assets')
@@ -225,39 +249,56 @@ def test_downloadable_not_found(client):
 @pytest.mark.parametrize('uri, default, expected, content_type', [
     ('', 'default', 'default', 'application/octet-stream'),
     ('other', 'default.html', 'default.html', 'text/html'),
+    ('zip', 'default.zip', 'default.zip', 'application/zip'),
     ('index2', 'index', 'index2', 'application/octet-stream'),
     ('absolute', '/foo/bar/index', '/foo/bar/index', 'application/octet-stream'),
     ('docs/notes/test.txt', 'index.html', 'index.html', 'text/html'),
     ('index.html_files/test.txt', 'index.html', 'index.html_files/test.txt', 'text/plain'),
 ])
 @pytest.mark.parametrize('downloadable', [True, False])
-def test_fallback_filename(uri, default, expected, content_type, downloadable,
+def test_fallback_filename(asgi, uri, default, expected, content_type, downloadable,
                            monkeypatch):
 
-    def mockOpen(path, mode):
-        if default in path:
-            return path
+    def mock_open(path, mode):
+        if os.path.normpath(default) in path:
+            return io.BytesIO(path.encode())
+
         raise IOError()
 
-    monkeypatch.setattr(io, 'open', mockOpen)
-    monkeypatch.setattr('os.path.isfile', lambda file: default in file)
+    monkeypatch.setattr(io, 'open', mock_open)
+    monkeypatch.setattr('os.path.isfile', lambda file: os.path.normpath(default) in file)
 
-    sr = StaticRoute('/static', '/var/www/statics', downloadable=downloadable,
-                     fallback_filename=default)
+    sr = create_sr(
+        asgi,
+        '/static',
+        '/var/www/statics',
+        downloadable=downloadable,
+        fallback_filename=default
+    )
 
     req_path = '/static/' + uri
 
-    req = Request(testing.create_environ(
+    req = _util.create_req(
+        asgi,
         host='test.com',
         path=req_path,
-        app='statics'
-    ))
-    resp = Response()
-    sr(req, resp)
+        root_path='statics'
+    )
+    resp = _util.create_resp(asgi)
+
+    if asgi:
+        async def run():
+            await sr(req, resp)
+            return await resp.stream.read()
+
+        body = falcon.async_to_sync(run)
+    else:
+        sr(req, resp)
+        body = resp.stream.read()
 
     assert sr.match(req.path)
-    assert resp.stream == os.path.join('/var/www/statics', expected)
-    assert resp.content_type == content_type
+    assert body.decode() == os.path.normpath(os.path.join('/var/www/statics', expected))
+    assert resp.content_type in _MIME_ALTERNATIVE.get(content_type, (content_type, ))
 
     if downloadable:
         assert os.path.basename(expected) in resp.downloadable_as
@@ -278,7 +319,7 @@ def test_e2e_fallback_filename(client, monkeypatch, strip_slash, path, fallback,
 
     def mockOpen(path, mode):
         if 'index' in path and 'raise' not in path:
-            return [path.encode('utf-8')]
+            return io.BytesIO(path.encode())
         raise IOError()
 
     monkeypatch.setattr(io, 'open', mockOpen)
@@ -295,7 +336,7 @@ def test_e2e_fallback_filename(client, monkeypatch, strip_slash, path, fallback,
             assert response.status == falcon.HTTP_404
         else:
             assert response.status == falcon.HTTP_200
-            assert response.text == directory + expected
+            assert response.text == os.path.normpath(directory + expected)
 
     test('/static', '/opt/somesite/static/', static_exp)
     test('/assets', '/opt/somesite/assets/', assert_axp)
@@ -311,9 +352,9 @@ def test_e2e_fallback_filename(client, monkeypatch, strip_slash, path, fallback,
     ('index2', '/staticfoo', False),
     ('index2', '/static/foo', True),
 ])
-def test_match(default, path, expected, monkeypatch):
+def test_match(asgi, default, path, expected, monkeypatch):
     monkeypatch.setattr('os.path.isfile', lambda file: True)
-    sr = StaticRoute('/static', '/var/www/statics', fallback_filename=default)
+    sr = create_sr(asgi, '/static', '/var/www/statics', fallback_filename=default)
 
     assert sr.match(path) == expected
 
@@ -323,7 +364,7 @@ def test_filesystem_traversal_fuse(client, monkeypatch):
     def suspicious_normpath(path):
         return 'assets/../../../../' + path
 
-    monkeypatch.setattr('os.path.normpath', suspicious_normpath)
     client.app.add_static_route('/static', '/etc/nginx/includes/static-data')
+    monkeypatch.setattr('os.path.normpath', suspicious_normpath)
     response = client.simulate_request(path='/static/shadow')
     assert response.status == falcon.HTTP_404
