@@ -4,9 +4,9 @@ import falcon
 from falcon import routing
 
 
-class TestUriTemplates(object):
+class TestUriTemplates:
 
-    @pytest.mark.parametrize('value', (42, falcon.API))
+    @pytest.mark.parametrize('value', (42, falcon.App))
     def test_string_type_required(self, value):
         with pytest.raises(TypeError):
             routing.compile_uri_template(value)
