@@ -16,7 +16,7 @@ _SIZE_1_KB = 1024
 
 
 @pytest.mark.usefixtures('_setup_wsgi_server')
-class TestWSGIServer(object):
+class TestWSGIServer:
 
     def test_get(self):
         resp = requests.get(_SERVER_BASE_URL)
@@ -56,15 +56,15 @@ class TestWSGIServer(object):
         assert not resp.text
 
 
-def _run_server(stop_event):
-    class Things(object):
+def _run_server(stop_event, host, port):
+    class Things:
         def on_get(self, req, resp):
-            resp.body = req.remote_addr
+            resp.text = req.remote_addr
 
         def on_post(self, req, resp):
             # NOTE(kgriffs): Elsewhere we just use req.bounded_stream, so
             # here we read the stream directly to test that use case.
-            resp.body = req.stream.read(req.content_length or 0)
+            resp.text = req.stream.read(req.content_length or 0)
 
         def on_put(self, req, resp):
             # NOTE(kgriffs): Test that reading past the end does
@@ -72,9 +72,9 @@ def _run_server(stop_event):
             req_body = (req.bounded_stream.read(1)
                         for i in range(req.content_length + 1))
 
-            resp.body = b''.join(req_body)
+            resp.text = b''.join(req_body)
 
-    class Bucket(object):
+    class Bucket:
         def on_post(self, req, resp):
             # NOTE(kgriffs): This would normally block when
             # Content-Length is 0 and the WSGI input object.
@@ -82,34 +82,45 @@ def _run_server(stop_event):
             # make sure req.bounded_stream is what we think it is;
             # BoundedStream itself has its own unit tests in
             # test_request_body.py
-            resp.body = req.bounded_stream.read()
+            resp.text = req.bounded_stream.read()
 
             # NOTE(kgriffs): No need to also test the same read() for
             # req.stream, since we already asserted they are the same
             # objects.
 
-    api = application = falcon.API()
+    api = application = falcon.App()
     api.add_route('/', Things())
     api.add_route('/bucket', Bucket())
 
-    server = make_server(_SERVER_HOST, _SERVER_PORT, application)
+    server = make_server(host, port, application)
 
     while not stop_event.is_set():
         server.handle_request()
 
 
-@pytest.fixture
+@pytest.fixture(scope='module')
 def _setup_wsgi_server():
     stop_event = multiprocessing.Event()
     process = multiprocessing.Process(
         target=_run_server,
-        args=(stop_event,)
+        daemon=True,
+
+        # NOTE(kgriffs): Pass these explicitly since if multiprocessing is
+        #   using the 'spawn' start method, we can't depend on closures.
+        args=(stop_event, _SERVER_HOST, _SERVER_PORT)
     )
 
     process.start()
 
-    # NOTE(kgriffs): Let the server start up
-    time.sleep(0.2)
+    # NOTE(vytas): Give the server some time to start.
+    for attempt in range(3):
+        try:
+            requests.get(_SERVER_BASE_URL, timeout=1)
+            break
+        except requests.exceptions.RequestException:
+            pass
+
+        time.sleep(attempt + 0.2)
 
     yield
 
