@@ -4,14 +4,13 @@ import pytest
 
 import falcon
 from falcon import testing
-from falcon.util import compat
 
-unicode_message = u'Unicode: \x80'
+unicode_message = 'Unicode: \x80'
 
 
 @pytest.fixture
 def client():
-    app = falcon.API()
+    app = falcon.App()
 
     tehlogger = LoggerResource()
     app.add_route('/logger', tehlogger)
@@ -27,20 +26,15 @@ class LoggerResource:
         req.log_error(unicode_message.encode('utf-8'))
 
 
-class TestWSGIError(object):
+class TestWSGIError:
 
     def setup_method(self, method):
         self.wsgierrors_buffer = io.BytesIO()
 
-        if compat.PY3:
-            # Simulate Gunicorn's behavior under Python 3
-            self.wsgierrors = io.TextIOWrapper(self.wsgierrors_buffer,
-                                               line_buffering=True,
-                                               encoding='utf-8')
-        else:
-            # WSGI servers typically present an open file object,
-            # with undefined encoding, so do the encoding manually.
-            self.wsgierrors = self.wsgierrors_buffer
+        # Simulate Gunicorn's behavior under Python 3
+        self.wsgierrors = io.TextIOWrapper(self.wsgierrors_buffer,
+                                           line_buffering=True,
+                                           encoding='utf-8')
 
     def test_responder_logged_bytestring(self, client):
         client.simulate_request(path='/logger',
@@ -51,11 +45,3 @@ class TestWSGIError(object):
 
         assert unicode_message.encode('utf-8') in log
         assert b'?amount=10' in log
-
-    @pytest.mark.skipif(compat.PY3, reason='Test only applies to Python 2')
-    def test_responder_logged_unicode(self, client):
-        client.simulate_request(path='/logger',
-                                wsgierrors=self.wsgierrors,
-                                method='HEAD')
-        log = self.wsgierrors_buffer.getvalue()
-        assert unicode_message in log.decode('utf-8')
