@@ -5,7 +5,7 @@ import falcon
 import falcon.testing as testing
 
 
-class TestWSGIInterface(object):
+class TestWSGIInterface:
 
     def test_srmock(self):
         mock = testing.StartResponseMock()
@@ -21,7 +21,7 @@ class TestWSGIInterface(object):
         assert mock.exc_info == exc_info
 
     def test_pep3333(self):
-        api = falcon.API()
+        api = falcon.App()
         mock = testing.StartResponseMock()
 
         # Simulate a web request (normally done though a WSGI server)
