@@ -11,17 +11,17 @@ class TypeResource(testing.SimpleTestResource):
         resp.status = falcon.HTTP_200
         # NOTE(masterkale): No size needs to be specified here because we're
         # emulating a stream read in production.
-        resp.body = json.dumps({'data': req.bounded_stream.read().decode('utf-8')})
+        resp.text = json.dumps({'data': req.bounded_stream.read().decode('utf-8')})
 
 
-class TestWsgiRefInputWrapper(object):
+class TestWsgiRefInputWrapper:
     def test_resources_can_read_request_stream_during_tests(self):
         """Make sure we can perform a simple request during testing.
 
         Originally, testing would fail after performing a request because no
         size was specified when calling `wsgiref.validate.InputWrapper.read()`
         via `req.stream.read()`"""
-        app = falcon.API()
+        app = falcon.App()
         type_route = '/type'
         app.add_route(type_route, TypeResource())
         client = testing.TestClient(app)
